# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import re
from xml.etree import ElementTree

import pytest

import psrpcore
from psrpcore.types import PipelineResultTypes, PSObject, SessionCapability

from .conftest import assert_xml_diff, deserialize, serialize


@pytest.mark.parametrize(
    "kwds, expected",
    [
        ({"name": "name"}, "command_text='name' is_script=False use_local_scope=None end_of_statement=False"),
        (
            {"name": "name", "is_script": True},
            "command_text='name' is_script=True use_local_scope=None end_of_statement=False",
        ),
        (
            {"name": "name", "use_local_scope": False},
            "command_text='name' is_script=False use_local_scope=False end_of_statement=False",
        ),
    ],
    ids=["name only", "is_script", "use_local_scope"],
)
def test_command_repr(kwds, expected):
    cmd = psrpcore.Command(**kwds)
    assert repr(cmd) == f"<Command {expected}>"
    assert str(cmd) == kwds["name"]


def test_add_argument():
    cmd = psrpcore.Command("cmd")
    cmd.add_argument("str").add_argument(1)

    element = serialize(cmd)
    actual = ElementTree.tostring(element, method="xml", encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        "<MS>"
        '<S N="Cmd">cmd</S>'
        '<Obj RefId="1" N="Args">'
        '<TN RefId="0">'
        "<T>System.Collections.ArrayList</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<LST>"
        '<Obj RefId="2"><MS><Nil N="N" /><S N="V">str</S></MS></Obj>'
        '<Obj RefId="3">'
        "<MS>"
        '<Nil N="N" />'
        '<I32 N="V">1</I32>'
        "</MS>"
        "</Obj>"
        "</LST>"
        "</Obj>"
        '<B N="IsScript">false</B>'
        '<Nil N="UseLocalScope" />'
        '<Obj RefId="4" N="MergeMyResult">'
        "<I32>0</I32>"
        '<TN RefId="1">'
        "<T>System.Management.Automation.Runspaces.PipelineResultTypes</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>None</ToString>"
        "</Obj>"
        '<Ref RefId="4" N="MergeToResult" />'
        '<Ref RefId="4" N="MergePreviousResults" />'
        "</MS>"
        "<ToString>cmd</ToString>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    raw_cmd = deserialize(element)
    assert isinstance(raw_cmd, PSObject)
    assert not isinstance(raw_cmd, psrpcore.Command)

    cmd = psrpcore.Command.FromPSObjectForRemoting(raw_cmd)
    assert isinstance(cmd, psrpcore.Command)

    assert cmd.command_text == "cmd"
    assert not cmd.is_script
    assert cmd.use_local_scope is None
    assert cmd.parameters == [(None, "str"), (None, 1)]
    assert cmd.end_of_statement is False
    assert not cmd.merge_unclaimed
    assert cmd.merge_my == PipelineResultTypes.none
    assert cmd.merge_to == PipelineResultTypes.none
    assert cmd.merge_error == PipelineResultTypes.none
    assert cmd.merge_warning == PipelineResultTypes.none
    assert cmd.merge_verbose == PipelineResultTypes.none
    assert cmd.merge_debug == PipelineResultTypes.none
    assert cmd.merge_information == PipelineResultTypes.none


def test_add_parameter():
    cmd = psrpcore.Command("cmd")
    cmd.add_parameter("param1", "test").add_parameter("param2", True)
    cmd.redirect_error()
    cmd.redirect_information()

    element = serialize(
        cmd,
        their_capability=SessionCapability(
            PSVersion="2.0",
            protocolversion="2.3",
            SerializationVersion="1.1.0.1",
        ),
    )
    actual = ElementTree.tostring(element, method="xml", encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        "<MS>"
        '<S N="Cmd">cmd</S>'
        '<Obj RefId="1" N="Args">'
        '<TN RefId="0">'
        "<T>System.Collections.ArrayList</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<LST>"
        '<Obj RefId="2"><MS><S N="N">param1</S><S N="V">test</S></MS></Obj>'
        '<Obj RefId="3"><MS><S N="N">param2</S><B N="V">true</B></MS></Obj>'
        "</LST>"
        "</Obj>"
        '<B N="IsScript">false</B>'
        '<Nil N="UseLocalScope" />'
        '<Obj RefId="4" N="MergeMyResult">'
        "<I32>2</I32>"
        '<TN RefId="1">'
        "<T>System.Management.Automation.Runspaces.PipelineResultTypes</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>Error</ToString>"
        "</Obj>"
        '<Obj RefId="5" N="MergeToResult">'
        "<I32>1</I32>"
        '<TNRef RefId="1" />'
        "<ToString>Output</ToString>"
        "</Obj>"
        '<Obj RefId="6" N="MergePreviousResults">'
        "<I32>0</I32>"
        '<TNRef RefId="1" />'
        "<ToString>None</ToString>"
        "</Obj>"
        '<Ref RefId="5" N="MergeError" />'
        '<Ref RefId="6" N="MergeWarning" />'
        '<Ref RefId="6" N="MergeVerbose" />'
        '<Ref RefId="6" N="MergeDebug" />'
        '<Ref RefId="5" N="MergeInformation" />'
        "</MS>"
        "<ToString>cmd</ToString>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    raw_cmd = deserialize(element)
    assert isinstance(raw_cmd, PSObject)
    assert not isinstance(raw_cmd, psrpcore.Command)

    cmd = psrpcore.Command.FromPSObjectForRemoting(raw_cmd)
    assert isinstance(cmd, psrpcore.Command)

    assert cmd.command_text == "cmd"
    assert not cmd.is_script
    assert cmd.use_local_scope is None
    assert cmd.parameters == [("param1", "test"), ("param2", True)]
    assert cmd.end_of_statement is False
    assert not cmd.merge_unclaimed
    assert cmd.merge_my == PipelineResultTypes.Error
    assert cmd.merge_to == PipelineResultTypes.Output
    assert cmd.merge_error == PipelineResultTypes.Output
    assert cmd.merge_warning == PipelineResultTypes.none
    assert cmd.merge_verbose == PipelineResultTypes.none
    assert cmd.merge_debug == PipelineResultTypes.none
    assert cmd.merge_information == PipelineResultTypes.Output


def test_add_parameters():
    cmd = psrpcore.Command("cmd")
    cmd.add_parameters(param1="test", param2=True)
    cmd.redirect_error(PipelineResultTypes.Null)
    cmd.redirect_information(PipelineResultTypes.Null)

    element = serialize(
        cmd,
        their_capability=SessionCapability(
            PSVersion="2.0",
            protocolversion="2.2",
            SerializationVersion="1.1.0.1",
        ),
    )
    actual = ElementTree.tostring(element, method="xml", encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        "<MS>"
        '<S N="Cmd">cmd</S>'
        '<Obj RefId="1" N="Args">'
        '<TN RefId="0">'
        "<T>System.Collections.ArrayList</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<LST>"
        '<Obj RefId="2"><MS><S N="N">param1</S><S N="V">test</S></MS></Obj>'
        '<Obj RefId="3"><MS><S N="N">param2</S><B N="V">true</B></MS></Obj>'
        "</LST>"
        "</Obj>"
        '<B N="IsScript">false</B>'
        '<Nil N="UseLocalScope" />'
        '<Obj RefId="4" N="MergeMyResult">'
        "<I32>0</I32>"
        '<TN RefId="1">'
        "<T>System.Management.Automation.Runspaces.PipelineResultTypes</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>None</ToString>"
        "</Obj>"
        '<Ref RefId="4" N="MergeToResult" />'
        '<Ref RefId="4" N="MergePreviousResults" />'
        '<Obj RefId="5" N="MergeError">'
        "<I32>8</I32>"
        '<TNRef RefId="1" />'
        "<ToString>Null</ToString>"
        "</Obj>"
        '<Ref RefId="4" N="MergeWarning" />'
        '<Ref RefId="4" N="MergeVerbose" />'
        '<Ref RefId="4" N="MergeDebug" />'
        "</MS>"
        "<ToString>cmd</ToString>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    raw_cmd = deserialize(element)
    assert isinstance(raw_cmd, PSObject)
    assert not isinstance(raw_cmd, psrpcore.Command)

    cmd = psrpcore.Command.FromPSObjectForRemoting(raw_cmd)
    assert isinstance(cmd, psrpcore.Command)

    assert cmd.command_text == "cmd"
    assert not cmd.is_script
    assert cmd.use_local_scope is None
    assert cmd.parameters == [("param1", "test"), ("param2", True)]
    assert cmd.end_of_statement is False
    assert not cmd.merge_unclaimed
    assert cmd.merge_my == PipelineResultTypes.none
    assert cmd.merge_to == PipelineResultTypes.none
    assert cmd.merge_error == PipelineResultTypes.Null
    assert cmd.merge_warning == PipelineResultTypes.none
    assert cmd.merge_verbose == PipelineResultTypes.none
    assert cmd.merge_debug == PipelineResultTypes.none
    assert cmd.merge_information == PipelineResultTypes.none


def test_merge_unclaimed():
    cmd = psrpcore.Command("test")
    cmd.merge_unclaimed = True

    element = serialize(cmd)
    actual = ElementTree.tostring(element, method="xml", encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        "<MS>"
        '<S N="Cmd">test</S>'
        '<Obj RefId="1" N="Args">'
        '<TN RefId="0">'
        "<T>System.Collections.ArrayList</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<LST />"
        "</Obj>"
        '<B N="IsScript">false</B>'
        '<Nil N="UseLocalScope" />'
        '<Obj RefId="2" N="MergeMyResult">'
        "<I32>0</I32>"
        '<TN RefId="1">'
        "<T>System.Management.Automation.Runspaces.PipelineResultTypes</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>None</ToString>"
        "</Obj>"
        '<Ref RefId="2" N="MergeToResult" />'
        '<Obj RefId="3" N="MergePreviousResults">'
        "<I32>3</I32>"
        '<TNRef RefId="1" />'
        "<ToString>Output, Error, Warning</ToString>"
        "</Obj>"
        "</MS>"
        "<ToString>test</ToString>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    raw_cmd = deserialize(element)
    assert isinstance(raw_cmd, PSObject)
    assert not isinstance(raw_cmd, psrpcore.Command)

    cmd = psrpcore.Command.FromPSObjectForRemoting(raw_cmd)
    assert isinstance(cmd, psrpcore.Command)
    assert cmd.merge_unclaimed


def test_redirect_error():
    cmd = psrpcore.Command("test")

    cmd.redirect_all(PipelineResultTypes.Null)
    assert cmd.merge_my == PipelineResultTypes.none
    assert cmd.merge_to == PipelineResultTypes.none
    assert cmd.merge_error == PipelineResultTypes.Null
    assert cmd.merge_warning == PipelineResultTypes.Null
    assert cmd.merge_verbose == PipelineResultTypes.Null
    assert cmd.merge_verbose == PipelineResultTypes.Null
    assert cmd.merge_information == PipelineResultTypes.Null

    cmd.redirect_all(PipelineResultTypes.Output)
    assert cmd.merge_my == PipelineResultTypes.Error
    assert cmd.merge_to == PipelineResultTypes.Output
    assert cmd.merge_error == PipelineResultTypes.Output
    assert cmd.merge_warning == PipelineResultTypes.Output
    assert cmd.merge_verbose == PipelineResultTypes.Output
    assert cmd.merge_verbose == PipelineResultTypes.Output
    assert cmd.merge_information == PipelineResultTypes.Output

    cmd.redirect_all(PipelineResultTypes.none)
    assert cmd.merge_my == PipelineResultTypes.none
    assert cmd.merge_to == PipelineResultTypes.none
    assert cmd.merge_error == PipelineResultTypes.none
    assert cmd.merge_warning == PipelineResultTypes.none
    assert cmd.merge_verbose == PipelineResultTypes.none
    assert cmd.merge_verbose == PipelineResultTypes.none
    assert cmd.merge_information == PipelineResultTypes.none


def test_redirect_to_invalid_output():
    expected = re.escape("Invalid redirection stream, must be none, Output, or Null")

    with pytest.raises(ValueError, match=expected):
        psrpcore.Command("test").redirect_error(PipelineResultTypes.Verbose)



# -*- coding: utf-8 -*-
# Copyright: (c) 2021", Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""The .NET type code.

Contains the public entrypoint for all the .NET/Pwsh type code that is used to
exchange objects between Python and PowerShell.
"""

import logging

from psrpcore.types._base import (
    PSAliasProperty,
    PSCryptoProvider,
    PSNoteProperty,
    PSObject,
    PSPropertyInfo,
    PSScriptProperty,
    PSType,
    add_alias_property,
    add_member,
    add_note_property,
    add_script_property,
    ps_isinstance,
)
from psrpcore.types._collection import (
    PSDict,
    PSDictBase,
    PSIEnumerable,
    PSList,
    PSListBase,
    PSQueue,
    PSQueueBase,
    PSStack,
    PSStackBase,
)
from psrpcore.types._complex import (
    ApartmentState,
    CommandMetadataCount,
    CommandOrigin,
    CommandTypes,
    ConsoleColor,
    DebugRecord,
    ErrorCategory,
    ErrorCategoryInfo,
    ErrorDetails,
    ErrorRecord,
    InformationalRecord,
    InformationRecord,
    InvocationInfo,
    NETException,
    PipelineResultTypes,
    ProgressRecord,
    ProgressRecordType,
    PSCredential,
    PSCredentialTypes,
    PSCredentialUIOptions,
    PSCustomObject,
    PSInvocationState,
    PSPrimitiveDictionary,
    PSThreadOptions,
    RemoteStreamOptions,
    RunspacePoolState,
    ScriptExtent,
    ScriptPosition,
    SessionStateEntryVisibility,
    VerboseRecord,
    WarningRecord,
)
from psrpcore.types._enum import PSEnumBase, PSFlagBase
from psrpcore.types._host import (
    BufferCell,
    BufferCellType,
    ChoiceDescription,
    ControlKeyStates,
    Coordinates,
    FieldDescription,
    HostDefaultData,
    HostInfo,
    HostMethodIdentifier,
    KeyInfo,
    ReadKeyOptions,
    Rectangle,
    Size,
)
from psrpcore.types._primitive import (
    PSBool,
    PSByte,
    PSByteArray,
    PSChar,
    PSDateTime,
    PSDecimal,
    PSDouble,
    PSDuration,
    PSGuid,
    PSInt,
    PSInt16,
    PSInt64,
    PSNull,
    PSSByte,
    PSScriptBlock,
    PSSecureString,
    PSSingle,
    PSString,
    PSUInt,
    PSUInt16,
    PSUInt64,
    PSUri,
    PSVersion,
    PSXml,
)
from psrpcore.types._psrp import (
    ApplicationPrivateData,
    ConnectRunspacePool,
    CreatePipeline,
    DebugRecordMsg,
    EncryptedSessionKey,
    EndOfPipelineInput,
    ErrorRecordMsg,
    GetAvailableRunspaces,
    GetCommandMetadata,
    InformationRecordMsg,
    InitRunspacePool,
    PipelineHostCall,
    PipelineHostResponse,
    PipelineInput,
    PipelineOutput,
    PipelineState,
    ProgressRecordMsg,
    PSRPMessageType,
    PublicKey,
    PublicKeyRequest,
    ResetRunspaceState,
    RunspaceAvailability,
    RunspacePoolHostCall,
    RunspacePoolHostResponse,
    RunspacePoolInitData,
    RunspacePoolStateMsg,
    SessionCapability,
    SetMaxRunspaces,
    SetMinRunspaces,
    UserEvent,
    VerboseRecordMsg,
    WarningRecordMsg,
)
from psrpcore.types._serializer import deserialize, serialize

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "ApartmentState",
    "ApplicationPrivateData",
    "BufferCell",
    "BufferCellType",
    "ChoiceDescription",
    "CommandMetadataCount",
    "CommandOrigin",
    "CommandTypes",
    "ConnectRunspacePool",
    "ConsoleColor",
    "ControlKeyStates",
    "Coordinates",
    "CreatePipeline",
    "DebugRecord",
    "DebugRecordMsg",
    "EncryptedSessionKey",
    "EndOfPipelineInput",
    "ErrorCategory",
    "ErrorCategoryInfo",
    "ErrorDetails",
    "ErrorRecord",
    "ErrorRecordMsg",
    "FieldDescription",
    "GetAvailableRunspaces",
    "GetCommandMetadata",
    "HostDefaultData",
    "HostInfo",
    "HostMethodIdentifier",
    "InformationRecord",
    "InformationRecordMsg",
    "InformationalRecord",
    "InitRunspacePool",
    "InvocationInfo",
    "KeyInfo",
    "NETException",
    "PSAliasProperty",
    "PSBool",
    "PSByte",
    "PSByteArray",
    "PSChar",
    "PSCredential",
    "PSCredentialTypes",
    "PSCredentialUIOptions",
    "PSCryptoProvider",
    "PSCustomObject",
    "PSDateTime",
    "PSDecimal",
    "PSDict",
    "PSDictBase",
    "PSDouble",
    "PSDuration",
    "PSEnumBase",
    "PSFlagBase",
    "PSGuid",
    "PSIEnumerable",
    "PSInt",
    "PSInt16",
    "PSInt64",
    "PSInvocationState",
    "PSList",
    "PSListBase",
    "PSNoteProperty",
    "PSNull",
    "PSObject",
    "PSPrimitiveDictionary",
    "PSPropertyInfo",
    "PSQueue",
    "PSQueueBase",
    "PSRPMessageType",
    "PSSByte",
    "PSScriptBlock",
    "PSScriptProperty",
    "PSSecureString",
    "PSSingle",
    "PSStack",
    "PSStackBase",
    "PSString",
    "PSThreadOptions",
    "PSType",
    "PSUInt",
    "PSUInt16",
    "PSUInt64",
    "PSUri",
    "PSVersion",
    "PSXml",
    "PipelineHostCall",
    "PipelineHostResponse",
    "PipelineInput",
    "PipelineOutput",
    "PipelineResultTypes",
    "PipelineState",
    "ProgressRecord",
    "ProgressRecordMsg",
    "ProgressRecordType",
    "PublicKey",
    "PublicKeyRequest",
    "ReadKeyOptions",
    "Rectangle",
    "RemoteStreamOptions",
    "ResetRunspaceState",
    "RunspaceAvailability",
    "RunspacePoolHostCall",
    "RunspacePoolHostResponse",
    "RunspacePoolInitData",
    "RunspacePoolState",
    "RunspacePoolStateMsg",
    "ScriptExtent",
    "ScriptPosition",
    "SessionCapability",
    "SessionStateEntryVisibility",
    "SetMaxRunspaces",
    "SetMinRunspaces",
    "Size",
    "UserEvent",
    "VerboseRecord",
    "VerboseRecordMsg",
    "WarningRecord",
    "WarningRecordMsg",
    "add_alias_property",
    "add_member",
    "add_note_property",
    "add_script_property",
    "deserialize",
    "ps_isinstance",
    "serialize",
]



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import logging
import typing
import uuid
from xml.etree import ElementTree

from psrpcore._crypto import PSRemotingCrypto
from psrpcore._events import PSRPEvent, SessionCapabilityEvent
from psrpcore._exceptions import (
    InvalidPipelineState,
    InvalidRunspacePoolState,
    PSRPCoreError,
)
from psrpcore._payload import (
    ProtocolVersion,
    PSRPMessage,
    PSRPPayload,
    StreamType,
    create_fragment,
    create_message,
    unpack_fragment,
    unpack_message,
)
from psrpcore.types import (
    ApartmentState,
    EndOfPipelineInput,
    ErrorRecord,
    HostInfo,
    PSInvocationState,
    PSObject,
    PSRPMessageType,
    PSThreadOptions,
    PSVersion,
    RunspacePoolState,
    SessionCapability,
    deserialize,
    serialize,
)

log = logging.getLogger(__name__)

T1 = typing.TypeVar("T1", bound="Pipeline")
T2 = typing.TypeVar("T2", bound="RunspacePool")


class RunspacePool(typing.Generic[T1]):
    """Runspace Pool base class.

    This is the base class for a Runspace Pool. It contains the common
    attributes and methods used by both a client and server based Runspace
    Pool.

    Args:
        runspace_pool_id: The UUID that identified the Runspace Pool.
        application_arguments: Any arguments supplied when creating the
            Runspace Pool as a client.
        application_private_data: Any special data supplied by the Runspace
            Pool as a server.
        ps_version: The PowerShell version.
        protocol_version: The PSRP protocol version that the pool understands.
        serialization_version: The serialization version used by the pool.

    Attributes:
        runspace_pool_id: See args.
        our_capability: The SessionCapability of the caller.
        their_capability: The SessionCapability of the peer, only populated
            after the Runspace Pool has been opened.
        application_arguments: The application arguments from the client, will
            be populated for the server after the Runspace Pool has been
            opened.
        application_private_data: The app private data supplied by the server,
            will be populated for the client after the Runspace Pool has been
            opened.
        host: The HostInfo that contains host information of the client.
        state: The current state of the Runspace Pool.
        apartment_state: The apartment state of the thread used to execute
            commands within this Runspace Pool.
        thread_options: Determines whether a new thread is created for each
            invocation.
        pipeline_table: A dictionary that contains associated pipelines with
            this Runspace Pool.
    """

    def __new__(
        cls,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> "RunspacePool":
        if cls == RunspacePool:
            raise TypeError(
                f"Type {cls.__qualname__} cannot be instantiated; it can be used only as a base class for "
                f"client/server runspace pool types."
            )

        return super().__new__(cls)

    def __init__(
        self,
        runspace_pool_id: uuid.UUID,
        application_arguments: typing.Dict[str, typing.Any],
        application_private_data: typing.Dict[str, typing.Any],
        ps_version: typing.Optional[PSVersion] = None,
        protocol_version: typing.Optional[PSVersion] = None,
        serialization_version: typing.Optional[PSVersion] = None,
    ) -> None:
        log.debug(
            "Creating Runspace Pool - RPID: %s, Type: %s",
            runspace_pool_id,
            type(self).__name__,
        )

        self.runspace_pool_id = runspace_pool_id
        self.our_capability = SessionCapability(
            PSVersion=ps_version or PSVersion("2.0"),
            protocolversion=protocol_version or ProtocolVersion.Pwsh5.value,
            SerializationVersion=serialization_version or PSVersion("1.1.0.1"),
        )
        self.their_capability: typing.Optional[SessionCapability] = None
        self.application_arguments = application_arguments
        self.application_private_data = application_private_data
        self.host: typing.Optional[HostInfo] = None
        self.state = RunspacePoolState.BeforeOpen
        self.apartment_state = ApartmentState.Unknown
        self.thread_options = PSThreadOptions.Default
        self.pipeline_table: typing.Dict[uuid.UUID, T1] = {}

        self._is_client = True
        self._ci_handlers: typing.Dict[int, typing.Optional[typing.Callable[[PSRPEvent], None]]] = {}
        self._ci_events: typing.Dict[int, PSRPEvent] = {}
        self._ci_count = 1
        self._fragment_count = 1
        self._key_requested = False
        self._cipher = PSRemotingCrypto()
        self._min_runspaces = 0
        self._max_runspaces = 0
        self._send_buffer: typing.List[PSRPMessage] = []

        # Raw bytes received but not yet processed.
        self._receive_buffer = bytearray()

        # Fragments for each object_id that have been received but not yet combined to a message.
        self._incoming_fragments: typing.Dict[int, typing.List[bytearray]] = {}

        # Messages from combined fragments that have been received but not yet returned.
        self._incoming_messages: typing.Dict[int, PSRPMessage] = {}

    @property
    def max_runspaces(
        self,
    ) -> int:
        """The maximum number of runspaces the pool maintains."""
        return self._max_runspaces

    @property
    def min_runspaces(
        self,
    ) -> int:
        """The minimum number of runspaces the pool maintains."""
        return self._min_runspaces

    @property
    def _ci_counter(
        self,
    ) -> int:
        """Counter used for ci calls."""
        ci = self._ci_count
        self._ci_count += 1
        return ci

    @property
    def _fragment_counter(
        self,
    ) -> int:
        """Counter used for fragment object IDs."""
        count = self._fragment_count
        self._fragment_count += 1
        return count

    def begin_close(self) -> None:
        """Marks the Runspace Pool to be in the closing phase."""
        self._change_state(RunspacePoolState.Closing)

    def close(self) -> None:
        """Marks the Runspace Pool as closed.

        This closes the Runspace Pool. Communicating to the peer that the pool
        is being closed is done through a connection specific process. This
        method just verifies the Runspace Pool is in a state that can be closed
        and that no pipelines are still running.
        """
        if self.pipeline_table:
            raise PSRPCoreError("Must close existing pipelines before closing the pool")

        valid_states = [RunspacePoolState.Closed, RunspacePoolState.Closing, RunspacePoolState.Opened]
        if self.state not in valid_states:
            raise InvalidRunspacePoolState("close Runspace Pool", self.state, valid_states)

        self._change_state(RunspacePoolState.Closed)

    def begin_disconnect(self) -> None:
        """Marks the Runspace Pool to be in the disconnecting phase."""
        self._change_state(RunspacePoolState.Disconnecting)

    def disconnect(self) -> None:
        """Marks the Runspace Pool as disconnected.

        This disconnects the Runspace Pool. COmmunicating to the peer that the
        pool is disconnected is done through a connection specific process.
        """
        valid_states = [RunspacePoolState.Opened, RunspacePoolState.Disconnecting, RunspacePoolState.Disconnected]
        if self.state not in valid_states:
            raise InvalidRunspacePoolState("disconnect a Runspace Pool", self.state, valid_states)

        self._change_state(RunspacePoolState.Disconnected)

    def reconnect(self) -> None:
        """Marks the Runspace Pool as reconnected and opened."""
        valid_states = [RunspacePoolState.Disconnected, RunspacePoolState.Opened]
        if self.state not in valid_states:
            raise InvalidRunspacePoolState("reconnect to a Runspace Pool", self.state, valid_states)

        self._change_state(RunspacePoolState.Opened)

    def data_to_send(
        self,
        amount: typing.Optional[int] = None,
    ) -> typing.Optional[PSRPPayload]:
        """Gets the next PSRP payload.

        Returns the PSRPPayload that contains the data that needs to be sent
        to the peer. This is a non-blocking call and is used by the
        implementer to get the next PSRP payload that is then sent over it's
        transport.

        Args:
            amount: The maximum size of the data fragment that can be sent.
                This must be 22 or larger to fit the fragment headers.

        Returns:
             typing.Optional[PSRPPayload]: The payload (if any) that needs to
                be sent to the peer.
        """
        if amount is not None and amount < 22:
            raise ValueError("amount must be 22 or larger to fit a PSRP fragment")

        current_buffer = bytearray()
        stream_type = StreamType.default
        pipeline_id = None
        fragment_size = 21
        # TODO: prioritise prompt_response over default if the last fragment was an end fragment.

        for message in list(self._send_buffer):
            if amount is not None and amount < fragment_size:
                break

            if not current_buffer:
                stream_type = message.stream_type
                pipeline_id = message.pipeline_id

            # We can only combine fragments if they are for the same target.
            if pipeline_id != message.pipeline_id:
                break

            if amount is None:
                allowed_length = len(message)
            else:
                allowed_length = amount - fragment_size
                amount -= fragment_size + len(message)

            fragment = message.fragment(allowed_length)
            log.debug(
                "Packing fragment - OID: %s, FID: %s, Start: %s, End: %s, Length: %s",
                fragment.object_id,
                fragment.fragment_id,
                fragment.start,
                fragment.end,
                len(fragment.data),
            )

            current_buffer += create_fragment(fragment.object_id, fragment.fragment_id, fragment.data, fragment.end)

            if len(message) == 0:
                self._send_buffer.remove(message)

        return PSRPPayload(bytes(current_buffer), stream_type, pipeline_id) if current_buffer else None

    def receive_data(
        self,
        data: PSRPPayload,
    ) -> None:
        """Store any incoming data.

        Stores any incoming payloads in an internal buffer to be processed.
        This buffer is read when calling :meth:`next_event()`.

        Args:
            data: The PSRP payload data received from the transport.
        """
        self._receive_buffer += data.data

    def next_event(
        self,
    ) -> typing.Optional[PSRPEvent]:
        """Process data received from the peer.

        This processes any PSRP data that has been received from the peer. Will
        return the next PSRP event in the receive buffer or `None` if not
        enough data is available.

        Returns:
            typing.Optional[PSRPEvent]: The next event present in the incoming
                data buffer or `None` if not enough data has been received.
        """
        # First unpacks the raw receive buffer into messages.
        while self._receive_buffer:
            fragment = unpack_fragment(self._receive_buffer)
            log.debug(
                "Unpacked fragment - OID: %s, FID: %s, Start: %s, End: %s, Length: %s",
                fragment.object_id,
                fragment.fragment_id,
                fragment.start,
                fragment.end,
                len(fragment.data),
            )
            self._receive_buffer = self._receive_buffer[21 + len(fragment.data) :]

            buffer = self._incoming_fragments.setdefault(fragment.object_id, [])
            if fragment.fragment_id != len(buffer):
                raise PSRPCoreError(
                    f"Expecting fragment with a fragment id of {len(buffer)} not {fragment.fragment_id}"
                )
            buffer.append(fragment.data)

            if fragment.end:
                raw_message = unpack_message(bytearray(b"".join(buffer)))
                message = PSRPMessage(
                    raw_message.message_type, raw_message.data, raw_message.rpid, raw_message.pid, fragment.object_id
                )
                self._incoming_messages[fragment.object_id] = message
                del self._incoming_fragments[fragment.object_id]

        for object_id in list(self._incoming_messages.keys()):
            message = self._incoming_messages[object_id]
            # In case of a failure it is expected for the client to receive the correct data instead
            del self._incoming_messages[object_id]

            event = self._process_message(message)

            return event

        # Need more data from peer to produce an event.
        return None

    def prepare_message(
        self,
        message: typing.Optional[PSObject],
        message_type: typing.Optional[PSRPMessageType] = None,
        pipeline_id: typing.Optional[uuid.UUID] = None,
        stream_type: StreamType = StreamType.default,
    ) -> None:
        """Adds a PSRP message to send buffer.

        Adds the given PSRP message to the send buffer to be sent when the
        caller requires it to.

        Args:
            message: The PSObject to be send.
            message_type: Override the message type of the PSRP messae in case
                message is not an actual PSRP Message object.
            pipeline_id: The pipeline id the message is for or `None` if it
                targets the runspace pool.
            stream_type: The stream type the message is for.
        """
        if message is None:
            b_data = b""
            message_type = PSRPMessageType.PipelineOutput

        elif isinstance(message, EndOfPipelineInput):
            b_data = b""  # Special edge case for this particular message type
            message_type = PSRPMessageType.EndOfPipelineInput

        else:
            element = serialize(
                message,
                self._cipher,
                # Extra info we pass to the serializer that may adjust how objects are serialized.
                our_capability=self.our_capability,
                their_capability=self.their_capability,
            )
            b_data = ElementTree.tostring(element, encoding="utf-8", method="xml")

            if message_type is None:
                try:
                    message_type = PSRPMessageType.get_message_id(type(message))
                except KeyError:
                    raise ValueError("message_type must be specified when the message is not a PSRP message") from None

        b_msg = create_message(self._is_client, message_type, b_data, self.runspace_pool_id, pipeline_id)

        object_id = self._fragment_counter
        psrp_message = PSRPMessage(
            message_type, bytearray(b_msg), self.runspace_pool_id, pipeline_id, object_id, stream_type
        )

        log.debug(
            "Queuing PSRP Message - MessageType: %s,  RPID: %s,  PID: %s,  OID: %s, Type: %s - '%s'",
            message_type.name,
            self.runspace_pool_id,
            pipeline_id or "''",
            object_id,
            stream_type.name,
            b_data.decode(),
        )
        self._send_buffer.append(psrp_message)

    def _change_state(
        self,
        state: RunspacePoolState,
        error: typing.Optional[ErrorRecord] = None,
    ) -> None:
        log.debug("Runspace Pool State change %s -> %s", self.state.name, state.name)
        self.state = state

    def _process_message(
        self,
        message: PSRPMessage,
    ) -> PSRPEvent:
        """Process a TransportDataAction data message received from a peer."""
        data = message.data
        log.debug(
            "Processing PSRP Message - MessageType: %s,  RPID: %s,  PID: %s,  OID: %s, Type: %s - '%s'",
            message.message_type.name,
            message.runspace_pool_id,
            message.pipeline_id or "''",
            message.object_id,
            message.stream_type.name,
            data.decode(),
        )

        ps_object: typing.Any
        if message.message_type == PSRPMessageType.EndOfPipelineInput:
            # Special edge case for EndOfPipelineInput which has no data.
            ps_object = EndOfPipelineInput()

        elif message.message_type == PSRPMessageType.PipelineOutput and data == b"":
            ps_object = None

        else:
            ps_object = deserialize(
                ElementTree.fromstring(data),
                cipher=self._cipher,
                # Extra info we pass to the serializer that may adjust how objects are serialized.
                our_capability=self.our_capability,
                their_capability=self.their_capability,
            )

        event = PSRPEvent.create(message.message_type, ps_object, message.runspace_pool_id, message.pipeline_id)

        process_func = getattr(self, f"_process_{message.message_type.name}", None)
        if process_func:
            process_func(event)

        else:
            log.warning(f"Received {message.message_type!s} but could not process it")

        return event

    def _process_SessionCapability(
        self,
        event: SessionCapabilityEvent,
    ) -> None:
        # This is the only common message that is processed the same by clients and servers.
        self.their_capability = SessionCapability(
            PSVersion=event.ps_version,
            protocolversion=event.protocol_version,
            SerializationVersion=event.serialization_version,
        )


class Pipeline(typing.Generic[T2]):
    """Pipeline base class.

    This is the base class for a Pipeline. It contains the common attributes
    and methods used by both a client and server base Pipeline.

    Args:
        runspace_pool: The Runspace Pool the pipeline is part of. When
            initialised the pipeline will add itself to the Runspace Pool
            pipeline table.
        pipeline_id: The Pipeline identifier.

    Attributes:
        runspace_pool: See args.
        pipeline_id: See args.
        state: The pipeline state.
    """

    def __new__(
        cls,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> "Pipeline":
        if cls == Pipeline:
            raise TypeError(
                f"Type {cls.__qualname__} cannot be instantiated; it can be used only as a base class for "
                f"client/server pipeline types."
            )

        return super().__new__(cls)

    def __init__(
        self,
        runspace_pool: T2,
        pipeline_id: uuid.UUID,
    ) -> None:
        log.debug(
            "Creating pipeline - RPID: %s, PID: %s, Type: %s",
            runspace_pool.runspace_pool_id,
            pipeline_id,
            type(self).__name__,
        )
        self.runspace_pool = runspace_pool
        self.pipeline_id = pipeline_id
        self.state = PSInvocationState.NotStarted
        self.metadata: typing.Optional[PSObject] = None
        runspace_pool.pipeline_table[self.pipeline_id] = self

    def begin_stop(self) -> None:
        """Marks the Pipeline to be in the stopping phase."""
        valid_state = [PSInvocationState.Running, PSInvocationState.Stopping]
        if self.state not in valid_state:
            raise InvalidPipelineState("begin stopping a pipeline", self.state, valid_state)

        self._change_state(PSInvocationState.Stopping)

    def close(self) -> None:
        """Close the Pipeline.

        Closes the pipeline by removing itself from the Runspace Pool pipeline
        table.
        """
        valid_state = [
            PSInvocationState.NotStarted,
            PSInvocationState.Stopped,
            PSInvocationState.Stopping,
            PSInvocationState.Completed,
            PSInvocationState.Failed,
        ]
        if self.state not in valid_state:
            raise InvalidPipelineState("closing a pipeline", self.state, valid_state)
        self.runspace_pool.pipeline_table.pop(self.pipeline_id, None)

    def prepare_message(
        self,
        message: typing.Optional[PSObject],
        message_type: typing.Optional[PSRPMessageType] = None,
        stream_type: StreamType = StreamType.default,
    ) -> None:
        """Adds a PSRP message to send buffer.

        Adds the given PSRP message to the send buffer to be sent when the
        caller requires it to. This just calls `prepare_message` on the
        Runspace Pool but for this specific pipeline.

        Args:
            message: The PSObject to be send.
            message_type: Override the message type of the PSRP messae in case
                message is not an actual PSRP Message object.
            stream_type: The stream type the message is for.
        """
        self.runspace_pool.prepare_message(message, message_type, self.pipeline_id, stream_type)

    def _change_state(
        self,
        state: PSInvocationState,
        error: typing.Optional[ErrorRecord] = None,
    ) -> None:
        log.debug("Pipeline State change %s -> %s", self.state.name, state.name)
        self.state = state



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""PSRP Messages.

The PSRP message types that are used as part of the PSRP fragments exchanged
between the client and the server.

.. MS-PSRP 2.2.2 Message Types:
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/a6136495-3260-4ff9-a982-7edf7ec95af2
"""

import enum
import typing

from psrpcore.types._base import PSNoteProperty, PSObject, PSType, T, add_note_property
from psrpcore.types._collection import PSList
from psrpcore.types._complex import (
    ApartmentState,
    CommandTypes,
    DebugRecord,
    ErrorRecord,
    ProgressRecordType,
    PSPrimitiveDictionary,
    PSThreadOptions,
    RemoteStreamOptions,
    VerboseRecord,
    WarningRecord,
)
from psrpcore.types._host import HostInfo, HostMethodIdentifier
from psrpcore.types._primitive import (
    PSBool,
    PSByteArray,
    PSDateTime,
    PSGuid,
    PSInt,
    PSInt64,
    PSString,
    PSStringBase,
    PSUInt,
    PSVersion,
)


class PSRPMessageType(enum.IntEnum):
    """PSRP Mesage types.

    The PowerShell Remoting Protocol message types identifiers. Each value
    corresponds to a PSRP message type and the unique identifier for that type.
    """

    SessionCapability = 0x00010002  #: :class:`SessionCapability`
    InitRunspacePool = 0x00010004  #: :class:`InitRunspacePool`
    PublicKey = 0x00010005  #: :class:`PublicKey`
    EncryptedSessionKey = 0x00010006  #: :class:`EncryptedSessionKey`
    PublicKeyRequest = 0x00010007  #: :class:`PublicKeyRequest`
    SetMaxRunspaces = 0x00021002  #: :class:`SetMaxRunspaces`
    SetMinRunspaces = 0x00021003  #: :class:`SetMinRunspaces`
    RunspaceAvailability = 0x00021004  #: :class:`RunspaceAvailability`
    RunspacePoolState = 0x00021005  #: :class:`RunspacePoolState`
    CreatePipeline = 0x00021006  #: :class:`CreatePipeline`
    GetAvailableRunspaces = 0x00021007  #: :class:`GetAvailableRunspaces`
    UserEvent = 0x00021008  #: :class:`UserEvent`
    ApplicationPrivateData = 0x00021009  #: :class:`ApplicationPrivateData`
    GetCommandMetadata = 0x0002100A  #: :class:`GetCommandMetadata`
    RunspacePoolHostCall = 0x00021100  #: :class:`RunspacePoolHostCall`
    RunspacePoolHostResponse = 0x00021101  #: :class:`RunspacePoolHostResponse`
    PipelineInput = 0x00041002  #: :class:`PipelineInput`
    EndOfPipelineInput = 0x00041003  #: :class:`EndOfPipelineInput`
    PipelineOutput = 0x00041004  #: :class:`PipelineOutput`
    ErrorRecord = 0x00041005  #: :class:`ErrorRecordMsg`
    PipelineState = 0x00041006  #: :class:`PipelineState`
    DebugRecord = 0x00041007  #: :class:`DebugRecordMsg`
    VerboseRecord = 0x00041008  #: :class:`VerboseRecordMsg`
    WarningRecord = 0x00041009  #: :class:`WarningRecordMsg`
    ProgressRecord = 0x00041010  #: :class:`ProgressRecordMsg`
    InformationRecord = 0x00041011  #: :class:`InformationRecordMsg`
    PipelineHostCall = 0x00041100  #: :class:`PipelineHostCall`
    PipelineHostResponse = 0x00041101  #: :class:`PipelineHostResponse`
    ConnectRunspacePool = 0x00010008  #: :class:`ConnectRunspacePool`
    RunspacePoolInitData = 0x0002100B  #: :class:`RunspacePoolInitData`
    ResetRunspaceState = 0x0002100C  #: :class:`ResetRunspaceState`

    @classmethod
    def get_message_id(cls, msg: typing.Type[PSObject]) -> "PSRPMessageType":
        return _REGISTRY[msg]


_REGISTRY: typing.Dict[typing.Type[PSObject], PSRPMessageType] = {}


class PSMessageType(PSType):
    def __init__(
        self,
        message_type: PSRPMessageType,
        *args: typing.Any,
        skip_inheritance: bool = True,
        **kwargs: typing.Any,
    ):
        kwargs.pop("rehydrate", None)
        self.message_type = message_type

        super().__init__(*args, skip_inheritance=skip_inheritance, rehydrate=False, **kwargs)

    def __call__(
        self,
        cls: T,
    ) -> T:
        cls = super().__call__(cls)

        _REGISTRY[cls] = self.message_type

        return cls


@PSMessageType(
    PSRPMessageType.SessionCapability,
    extended_properties=[
        PSNoteProperty("PSVersion", mandatory=True, ps_type=PSVersion),
        PSNoteProperty("protocolversion", mandatory=True, ps_type=PSVersion),
        PSNoteProperty("SerializationVersion", mandatory=True, ps_type=PSVersion),
        PSNoteProperty("TimeZone", ps_type=PSByteArray),
    ],
)
class SessionCapability(PSObject):
    """SESSION_CAPABILITY Message.

    Defines the session capability and protocol versions. Message is defined in
    `MS-PSRP 2.2.2.1 SESSION_CAPABILITY`_. The TimeZone property is a .NET type
    serialized to bytes as defined by
    `MS-PSRP 2.2.3.10.1 CurrentSystemTimeZone`_.

    Args:
        PSVersion: The version of the higher-layer application.
        protocolversion: The version of the PowerShell Remoting Protocol.
        SerializationVersion: The version of the serialization system.
        TimeZone: The time zone of the client.

    .. _MS-PSRP 2.2.2.1 SESSION_CAPABILITY:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/2f41abfb-7e30-4fb1-b286-527e9d67ad30

    .. _MS-PSRP 2.2.3.10.1 CurrentSystemTimeZone:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/5e6263c5-358a-459b-a49e-0707e383eb55
    """

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "SessionCapability",
        **kwargs: typing.Any,
    ) -> PSObject:
        obj = PSObject()
        add_note_property(obj, "PSVersion", instance.PSVersion)
        add_note_property(obj, "protocolversion", instance.protocolversion)
        add_note_property(obj, "SerializationVersion", instance.SerializationVersion)

        if instance.TimeZone:
            add_note_property(obj, "TimeZone", instance.TimeZone)

        return obj


@PSMessageType(
    PSRPMessageType.InitRunspacePool,
    extended_properties=[
        PSNoteProperty("MinRunspaces", mandatory=True, ps_type=PSInt),
        PSNoteProperty("MaxRunspaces", mandatory=True, ps_type=PSInt),
        PSNoteProperty("PSThreadOptions", mandatory=True, ps_type=PSThreadOptions),
        PSNoteProperty("ApartmentState", mandatory=True, ps_type=ApartmentState),
        PSNoteProperty("HostInfo", mandatory=True, ps_type=HostInfo),
        PSNoteProperty("ApplicationArguments", mandatory=True, ps_type=PSPrimitiveDictionary),
    ],
)
class InitRunspacePool(PSObject):
    """INIT_RUNSPACEPOOL Message.

    Defines the Runspace Pool initialization data. Message is defined in
    `MS-PSRP 2.2.2.2 INIT_RUNSPACEPOOL`_.

    Args:
        MinRunspaces: The minimum number of runspaces in the Runspace Pool.
        MaxRunspaces: The maximum number of runspaces in the Runspace Pool.
        PSThreadOptions: Thread options provided by the higher layer.
        ApartmentState: Apart state provided by the higher layer.
        HostInfo: Host information.
        ApplicationArguments: Application arguments provided by the higher
            layer.

    .. _MS-PSRP 2.2.2.2 INIT_RUNSPACEPOOL:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/c867589a-0b43-47bd-9abf-7477699ff6c9

    """


@PSMessageType(
    PSRPMessageType.PublicKey,
    extended_properties=[
        PSNoteProperty("PublicKey", mandatory=True, ps_type=PSString),
    ],
)
class PublicKey(PSObject):
    """PUBLIC_KEY Message.

    Defines the public key created by the client used in the session key
    exchange. Message is defined in `MS-PSRP 2.2.2.3 - PUBLIC_KEY`_.

    Args:
        PublicKey: The base64 encoding of the PKCS1 formatted public key.

    .. _MS-PSRP 2.2.2.3 - PUBLIC_KEY:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/3efa4b90-c089-432b-91db-76a3deb175bc
    """


@PSMessageType(
    PSRPMessageType.EncryptedSessionKey,
    extended_properties=[
        PSNoteProperty("EncryptedSessionKey", mandatory=True, ps_type=PSString),
    ],
)
class EncryptedSessionKey(PSObject):
    """ENCRYPTED_SESSION_KEY Message.

    Defines the encrypted session key calculated by the server. The value is
    encrypted using the public key that was sent by the client in the
    :class:`PublicKey` message. Message is defined in
    `MS-PSRP 2.2.2.4 - ENCRYPTED_SESSION_KEY`_.

    Note:
        The session key is encrypted using the RSAES-PKCS-v1_5 encryption
        scheme.

    Args:
        EncryptedSessionKey: The base64 encoding of the 256-bit AES encrypted
            session key.

    .. _MS-PSRP 2.2.2.4 - ENCRYPTED_SESSION_KEY:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/e3e155af-b379-40cf-80c0-14a124145147
    """


@PSMessageType(PSRPMessageType.PublicKeyRequest)
class PublicKeyRequest(PSStringBase):
    """PUBLIC_KEY_REQUEST Message.

    This is a message that the server sends the client when it wants to start
    the encryption key exchange. Message is defined in
    `MS-PSRP 2.2.2.5 - PUBLIC_KEY_REQUEST`_.

    .. _MS-PSRP 2.2.2.5 - PUBLIC_KEY_REQUEST:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/9ff2857d-a7cb-4da6-81f1-65d08b3dbe63
    """


@PSMessageType(
    PSRPMessageType.SetMaxRunspaces,
    extended_properties=[
        PSNoteProperty("MaxRunspaces", mandatory=True, ps_type=PSInt),
        PSNoteProperty("ci", mandatory=True, ps_type=PSInt64),
    ],
)
class SetMaxRunspaces(PSObject):
    """SET_MAX_RUNSPACES Message.

    Set maximum runspaces in a RunspacePool. Message is defined in
    `MS-PSRP 2.2.2.6 - SET_MAX_RUNSPACES`_.

    Args:
        MaxRunspaces: The maximum runspaces in a pool.
        ci: The Call ID the message is related to.

    .. _MS-PSRP 2.2.2.6 - SET_MAX_RUNSPACES:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/92037046-043a-4962-8e7e-2d457249548b
    """


@PSMessageType(
    PSRPMessageType.SetMinRunspaces,
    extended_properties=[
        PSNoteProperty("MinRunspaces", mandatory=True, ps_type=PSInt),
        PSNoteProperty("ci", mandatory=True, ps_type=PSInt64),
    ],
)
class SetMinRunspaces(PSObject):
    """SET_MIN_RUNSPACES Message.

    Set minimum runspaces in a RunspacePool. Message is defined in
    `MS-PSRP 2.2.2.7 - SET_MIN_RUNSPACES`_.

    Args:
        MinRunspaces: The minimum runspaces in a pool.
        ci: The Call ID the message is related to.

    .. _MS-PSRP 2.2.2.7 - SET_MIN_RUNSPACES:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/2d425c82-ead1-4888-911a-b11f545ca441
    """


@PSMessageType(
    PSRPMessageType.RunspaceAvailability,
    extended_properties=[
        PSNoteProperty("SetMinMaxRunspacesResponse", mandatory=True),
        PSNoteProperty("ci", mandatory=True, ps_type=PSInt64),
    ],
)
class RunspaceAvailability(PSObject):
    """RUNSPACE_AVAILABILITY Message.

    A response to either set maximum runspaces or set minimum runspaces in a
    RunspacePool or request for available runspaces in a RunspacePool. Message
    is defined in `MS-PSRP 2.2.2.8 - RUNSPACE_AVAILABILITY`_.

    Args:
        SetMinMaxRunspacesResponse: A :class:`PSBool` for a set min/max response
            or a :class:`PSInt64` for a get available response.
        ci: The Call ID the message is related to.

    .. _MS-PSRP 2.2.2.8 - RUNSPACE_AVAILABILITY:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/bcab75d9-31a8-4fdc-a8c4-00f41e5985d2
    """


@PSMessageType(
    PSRPMessageType.RunspacePoolState,
    extended_properties=[
        PSNoteProperty("RunspaceState", mandatory=True, ps_type=PSInt),
        PSNoteProperty("ExceptionAsErrorRecord", ps_type=ErrorRecord),
    ],
)
class RunspacePoolStateMsg(PSObject):
    """RUNSPACEPOOL_STATE Message.

    Defines the state of the RunspacePool. Message is defined in
    `MS-PSRP 2.2.2.9 RUNSPACEPOOL_STATE`_. The raw RunspaceState values
    correlate to :class:`RunspacePoolState`.

    Args:
        RunspaceState: The RunspacePool state information as an integer.
        ExceptionAsErrorRecord: The optional error record associated with the
            RunspacePool error.

    .. _MS-PSRP 2.2.2.9 RUNSPACEPOOL_STATE:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/0a5d8ef3-3b2c-4e16-9f2c-16efdaf16925
    """

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "RunspacePoolStateMsg",
        **kwargs: typing.Any,
    ) -> PSObject:
        obj = PSObject()
        add_note_property(obj, "RunspaceState", instance.RunspaceState)
        if instance.ExceptionAsErrorRecord:
            add_note_property(obj, "ExceptionAsErrorRecord", instance.ExceptionAsErrorRecord)

        return obj


@PSMessageType(
    PSRPMessageType.CreatePipeline,
    extended_properties=[
        PSNoteProperty("NoInput", mandatory=True, ps_type=PSBool),
        PSNoteProperty("ApartmentState", mandatory=True, ps_type=ApartmentState),
        PSNoteProperty("RemoteStreamOptions", mandatory=True, ps_type=RemoteStreamOptions),
        PSNoteProperty("AddToHistory", mandatory=True, ps_type=PSBool),
        PSNoteProperty("HostInfo", mandatory=True, ps_type=HostInfo),
        PSNoteProperty("PowerShell", mandatory=True),
        PSNoteProperty("IsNested", mandatory=True, ps_type=PSBool),
    ],
)
class CreatePipeline(PSObject):
    """CREATE_PIPELINE Message.

    Creates a command pipeline and invoke it in the specified RunspacePool.
    Message is defined in `MS-PSRP 2.2.2.10 CREATE_PIPELINE`_.

    Args:
        NoInput: Whether the pipeline will take input.
        ApartmentState: Apartment state provided by the higher layer.
        RemoteStreamOptions: Stream options that indicate how an application
            must treat messages from the PowerShell streams.
        AddToHistory: Whether the higher layer is to add the pipeline to the
            history field of the runspace.
        HostInfo: The host information.
        PowerShell: The pipeline information to create.
        IsNested: Whether the higher layer is to run the pipeline in nested or
            steppable mode.

    .. _MS-PSRP 2.2.2.10 CREATE_PIPELINE:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/2cf8cccb-63ab-404a-82df-caef0c41717a
    """


@PSMessageType(
    PSRPMessageType.GetAvailableRunspaces,
    extended_properties=[
        PSNoteProperty("ci", mandatory=True, ps_type=PSInt64),
    ],
)
class GetAvailableRunspaces(PSObject):
    """GET_AVAILABLE_RUNSPACES Message.

    Get the number of available runspaces in a RunspacePool. Message is defined
    in `MS-PSRP 2.2.2.11 GET_AVAILABLE_RUNSPACES`_.

    Args:
        ci: The Call ID associated with this operation.

    .. _MS-PSRP 2.2.2.11 GET_AVAILABLE_RUNSPACES:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/3f4d5a5c-9e7f-4ea2-8fea-253ddd394638
    """


@PSMessageType(
    PSRPMessageType.UserEvent,
    extended_properties=[
        PSNoteProperty("EventIdentifier", ps_type=PSInt),
        PSNoteProperty("SourceIdentifier", ps_type=PSString),
        PSNoteProperty("TimeGenerated", ps_type=PSDateTime),
        PSNoteProperty("Sender"),
        PSNoteProperty("SourceArgs", mandatory=True),
        PSNoteProperty("MessageData"),
        PSNoteProperty("ComputerName", ps_type=PSString),
        PSNoteProperty("RunspaceId", ps_type=PSGuid),
    ],
)
class UserEvent(PSObject):
    """USER_EVENT Message.

    Report a user-defined event from a remote runspace. Message is defined in
    `MS-PSRP 2.2.2.12 USER_EVENT`_.

    Args:
        EventIdentifier: The event identifier.
        SourceIdentifier: The source identifier.
        TimeGenerated: The time when the event was generated.
        Sender: The sender of the event.
        SourceArgs: Event arguments.
        MessageData: Message data associated with the event.
        ComputerName: Name of the computer where the event was fired.
        RunspaceId: The ID of the runspace.

    .. _MS-PSRP 2.2.2.12 USER_EVENT:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/c5a79f22-715d-4221-ae4d-47c685197b3b
    """

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "UserEvent":
        return UserEvent(
            EventIdentifier=obj["PSEventArgs.EventIdentifier"],
            SourceIdentifier=obj["PSEventArgs.SourceIdentifier"],
            TimeGenerated=obj["PSEventArgs.TimeGenerated"],
            Sender=obj["PSEventArgs.Sender"],
            SourceArgs=obj["PSEventArgs.SourceArgs"],
            MessageData=obj["PSEventArgs.MessageData"],
            ComputerName=obj["PSEventArgs.ComputerName"],
            RunspaceId=obj["PSEventArgs.RunspaceId"],
        )

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "UserEvent",
        **kwargs: typing.Any,
    ) -> PSObject:
        obj = PSObject()

        add_note_property(obj, "PSEventArgs.EventIdentifier", instance.EventIdentifier, ps_type=PSInt)
        add_note_property(obj, "PSEventArgs.SourceIdentifier", instance.SourceIdentifier, ps_type=PSString)
        add_note_property(obj, "PSEventArgs.TimeGenerated", instance.TimeGenerated, ps_type=PSDateTime)
        add_note_property(obj, "PSEventArgs.Sender", instance.Sender)
        add_note_property(obj, "PSEventArgs.SourceArgs", instance.SourceArgs)
        add_note_property(obj, "PSEventArgs.MessageData", instance.MessageData)
        add_note_property(obj, "PSEventArgs.ComputerName", instance.ComputerName, ps_type=PSString)
        add_note_property(obj, "PSEventArgs.RunspaceId", instance.RunspaceId, ps_type=PSGuid)

        return obj


@PSMessageType(
    PSRPMessageType.ApplicationPrivateData,
    extended_properties=[
        PSNoteProperty("ApplicationPrivateData", mandatory=True, ps_type=PSPrimitiveDictionary),
    ],
)
class ApplicationPrivateData(PSObject):
    """APPLICATION_PRIVATE_DATA Message.

    Data private to the application using the PowerShell Remoting Protocol on
    the server and client, which is passed by the protocol without
    interpretation. Message is defined in
    `MS-PSRP 2.2.2.13 APPLICATION_PRIVATE_DATA`_.

    Args:
        ApplicationPrivateData: Private data that the higher layer provides to
            the server when a RunspacePool is created.

    .. _MS-PSRP 2.2.2.13 APPLICATION_PRIVATE_DATA:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/f0e105d4-4242-429f-b63b-a600111fb27e
    """


@PSMessageType(
    PSRPMessageType.GetCommandMetadata,
    extended_properties=[
        PSNoteProperty("Name", ps_type=PSList),
        PSNoteProperty("CommandType", ps_type=CommandTypes),
        PSNoteProperty("Namespace", ps_type=PSList),
        PSNoteProperty("ArgumentList", ps_type=PSList),
    ],
)
class GetCommandMetadata(PSObject):
    """GET_COMMAND_METADATA Message.

    Get command metadata for commands available in a RunspacePool. Message is
    defined in `MS-PSRP 2.2.2.14 GET_COMMAND_METADATA`_.

    Args:
        Name: List of wildcard patterns specifying the command names that the
            server SHOULD return.
        CommandType: The command types to search for.
        Namespace: The command namespaces containing the commands that the
            server SHOULD return.
        ArgumentList: Extra arguments passed to the higher layer above the
            PowerShell Remoting Protocol.

    .. _MS-PSRP 2.2.2.14 GET_COMMAND_METADATA:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/b634ddef-93a0-4d3b-9e63-a630d01f233a
    """


@PSMessageType(
    PSRPMessageType.RunspacePoolHostCall,
    extended_properties=[
        PSNoteProperty("ci", mandatory=True, ps_type=PSInt64),
        PSNoteProperty("mi", mandatory=True, ps_type=HostMethodIdentifier),
        PSNoteProperty("mp", mandatory=True, ps_type=PSList),
    ],
)
class RunspacePoolHostCall(PSObject):
    """RUNSPACEPOOL_HOST_CALL Message.

    Method call on the host associated with the RunspacePool on the server.
    Message is defined in `MS-PSRP 2.2.2.15 RUNSPACEPOOL_HOST_CALL`_.

    Args:
        ci: The Call ID associated with this operation.
        mi: The host method identifier.
        mp: The parameters for the method.

    .. _MS-PSRP 2.2.2.15 RUNSPACEPOOL_HOST_CALL:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/4623540b-4dd3-440e-a54b-e0fb87dd92c8
    """


@PSMessageType(
    PSRPMessageType.RunspacePoolHostResponse,
    extended_properties=[
        PSNoteProperty("ci", mandatory=True, ps_type=PSInt64),
        PSNoteProperty("mi", mandatory=True, ps_type=HostMethodIdentifier),
        PSNoteProperty("mr"),
        PSNoteProperty("me", ps_type=ErrorRecord),
    ],
)
class RunspacePoolHostResponse(PSObject):
    """RUNSPACEPOOL_HOST_RESPONSE Message.

    Response from a host call executed on the client RunspacePool's host.
    Message is defined in `MS-PSRP 2.2.2.16 RUNSPACEPOOL_HOST_RESPONSE`_.

    Args:
        ci: The Call ID associated with this operation.
        mi: The host method identifier.
        mr: The return value of the method.
        me: Exception thrown by a host method invocation.

    .. _MS-PSRP 2.2.2.16 RUNSPACEPOOL_HOST_RESPONSE:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/9bcdf122-ad6b-45c3-9960-68d22627cdb5
    """

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "RunspacePoolHostResponse",
        **kwargs: typing.Any,
    ) -> PSObject:
        obj = PSObject()
        add_note_property(obj, "ci", instance.ci)
        add_note_property(obj, "mi", instance.mi)

        if instance.mr is not None:
            add_note_property(obj, "mr", instance.mr)

        if instance.me:
            add_note_property(obj, "me", instance.me)

        return obj


@PSMessageType(PSRPMessageType.PipelineInput)
class PipelineInput(PSObject):
    """PIPELINE_INPUT Message.

    Input to a command pipeline on the server. Message is defined in
    `MS-PSRP 2.2.2.17 PIPELINE_INPUT`_. The actual object is the serialized
    object that is being sent as input.

    .. _MS-PSRP 2.2.2.17 PIPELINE_INPUT:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/2c08acdd-3443-48c2-bf87-8fe2808d96ea
    """


@PSMessageType(PSRPMessageType.EndOfPipelineInput)
class EndOfPipelineInput(PSObject):
    """END_OF_PIPELINE_INPUT Message.

    Close the input collection for the command pipeline on the server. Message
    is defined in `MS-PSRP 2.2.2.18 END_OF_PIPELINE_INPUT`_.

    .. _MS-PSRP 2.2.2.18 END_OF_PIPELINE_INPUT:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/e616e6fd-0241-4823-b415-7dfc247646f1
    """


@PSMessageType(PSRPMessageType.PipelineOutput)
class PipelineOutput(PSObject):
    """PIPELINE_OUTPUT Message.

    Output of a command pipeline on the server. Message is defined in
    `MS-PSRP 2.2.2.19 PIPELINE_OUTPUT`_. The actual object is the serialized
    object that is being outputted from the server.

    .. _MS-PSRP 2.2.2.19 PIPELINE_OUTPUT:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/3b2c1076-c435-4aef-bdfe-3179bc452723
    """


@PSMessageType(PSRPMessageType.ErrorRecord, skip_inheritance=False)
class ErrorRecordMsg(ErrorRecord):
    """ERROR_RECORD Message.

    Error record from a command pipeline on the server.. Message is defined in
    `MS-PSRP 2.2.2.20 ERROR_RECORD`_.

    .. _MS-PSRP 2.2.2.20 ERROR_RECORD:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/c527797a-d017-4755-8a81-9f58280a7135
    """


@PSMessageType(
    PSRPMessageType.PipelineState,
    extended_properties=[
        PSNoteProperty("PipelineState", mandatory=True, ps_type=PSInt),
        PSNoteProperty("ExceptionAsErrorRecord", ps_type=ErrorRecord),
    ],
)
class PipelineState(PSObject):
    """PIPELINE_STATE Message.

    State information of a command pipeline on the server. Message is defined
    in `MS-PSRP 2.2.2.21 PIPELINE_STATE`_. The raw PipelineState values
    correlate to :class:`PSInvocationState`.

    Args:
        PipelineState: State information of the command pipeline.
        ExceptionAsErrorRecord: The optional error record associated with the
            Pipeline state error.

    .. _MS-PSRP 2.2.2.21 PIPELINE_STATE:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/932f0c9d-845a-4883-8efd-b49a593578b8
    """

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "PipelineState",
        **kwargs: typing.Any,
    ) -> PSObject:
        obj = PSObject()
        add_note_property(obj, "PipelineState", instance.PipelineState)
        if instance.ExceptionAsErrorRecord:
            add_note_property(obj, "ExceptionAsErrorRecord", instance.ExceptionAsErrorRecord)

        return obj


@PSMessageType(PSRPMessageType.DebugRecord, skip_inheritance=False)
class DebugRecordMsg(DebugRecord):
    """DEBUG_RECORD Message.

    Debug record from a command pipeline on the server. Message is defined in
    `MS-PSRP 2.2.2.22 DEBUG_RECORD`_.

    .. _MS-PSRP 2.2.2.22 DEBUG_RECORD:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/43b4cb30-6b14-498b-9325-c60339838a22
    """


@PSMessageType(PSRPMessageType.VerboseRecord, skip_inheritance=False)
class VerboseRecordMsg(VerboseRecord):
    """VERBOSE_RECORD Message.

    Verbose record from a command pipeline on the server. Message is defined in
    `MS-PSRP 2.2.2.23 VERBOSE_RECORD`_.

    .. _MS-PSRP 2.2.2.23 VERBOSE_RECORD:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/f94b18f5-0bd4-4817-8184-eb72767cce94
    """


@PSMessageType(PSRPMessageType.WarningRecord, skip_inheritance=False)
class WarningRecordMsg(WarningRecord):
    """WARNING_RECORD Message.

    Warning record from a command pipeline on the server. Message is defined in
    `MS-PSRP 2.2.2.24 WARNING_RECORD`_.

    .. _MS-PSRP 2.2.2.24 WARNING_RECORD:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/31c10c51-b831-475c-ae62-603426e6a617
    """


# While this looks similar to the ProgressRecord type the PSRP message is different and cannot inherit from
# ProgressRecord.
@PSMessageType(
    PSRPMessageType.ProgressRecord,
    extended_properties=[
        PSNoteProperty("Activity", ps_type=PSString),
        PSNoteProperty("ActivityId", ps_type=PSInt),
        PSNoteProperty("StatusDescription", ps_type=PSString),
        PSNoteProperty("CurrentOperation", ps_type=PSString),
        PSNoteProperty("ParentActivityId", ps_type=PSInt),
        PSNoteProperty("PercentComplete", ps_type=PSInt),
        PSNoteProperty("Type", ps_type=ProgressRecordType),
        PSNoteProperty("SecondsRemaining", ps_type=PSInt),
    ],
)
class ProgressRecordMsg(PSObject):
    """PROGRESS_RECORD Message.

    Progress record from a command pipeline on the server. Message is defined
    in `MS-PSRP 2.2.2.25 PROGRESS_RECORD`_.

    Args:
        Activity: Description of the activity.
        ActivityId: Id of the activity, used as a key for the linking of
            subordinate activities.
        StatusDescription: Current status of the operation, e.g. "35 of 50
            items copied".
        CurrentOperation: Current operation of the many required to accomplish
            the activity, e.g. "copying foo.txt".
        ParentActivityId: Id of the activity for which this is a subordinate.
        PercentComplete: Percentage of total work for the activity that is
            completed.
        Type: Type of record represented by this instance.
        SecondsRemaining: Estimated time remaining until the activity is
            complete.

    .. _MS-PSRP 2.2.2.25 PROGRESS_RECORD:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/435ab824-1069-43eb-8146-7c50593a47ac
    """


# While this has the same props as InformationRecord it is serialized as extended props and without types.
@PSMessageType(
    PSRPMessageType.InformationRecord,
    extended_properties=[
        PSNoteProperty("MessageData"),
        PSNoteProperty("Source", ps_type=PSString),
        PSNoteProperty("TimeGenerated", ps_type=PSDateTime),
        PSNoteProperty("Tags", ps_type=PSList),
        PSNoteProperty("User", ps_type=PSString),
        PSNoteProperty("Computer", ps_type=PSString),
        PSNoteProperty("ProcessId", ps_type=PSUInt),
        PSNoteProperty("NativeThreadId", ps_type=PSUInt),
        PSNoteProperty("ManagedThreadId", ps_type=PSUInt),
    ],
)
class InformationRecordMsg(PSObject):
    """INFORMATION_RECORD Message.

    Information record from a command pipeline on the server. Message is
    defined in `MS-PSRP 2.2.2.26 INFORMATION_RECORD`_.

    Note:
        This message is only used in ProtocolVersion>=2.3 (PowerShell v5.1+).

    Args:
        MessageData: The message data for the informational record.
        Source: The source of the information record (script path, function
            name, etc.).
        TimeGenerated: The time the informational record was generated.
        Tags: The tags associated with this informational record.
        User: THe user that generated the informational record.
        Computer: THe computer that generated the informational record.
        ProcessId: The process that generated the informational record.
        NativeThreadId: The native thread that generated the informational
            record.
        ManagedThreadId: The managed thread that generated the informational
            record.

    .. _MS-PSRP 2.2.2.26 INFORMATION_RECORD:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/5a3ec5f0-4654-4d87-830c-d3e07c4717c9
    """


@PSMessageType(PSRPMessageType.PipelineHostCall, skip_inheritance=False)
class PipelineHostCall(RunspacePoolHostCall):
    """PIPELINE_HOST_CALL Message.

    Method call on the host associated with the pipeline invocation settings on
    the server. Message is defined in `MS-PSRP 2.2.2.27 PIPELINE_HOST_CALL`_.

    Args:
        ci: The Call ID associated with this operation.
        mi: The host method identifier.
        mp: The parameters for the method.

    .. _MS-PSRP 2.2.2.27 PIPELINE_HOST_CALL:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/16947dfb-99b5-461f-b556-dec1beb33da8
    """


@PSMessageType(PSRPMessageType.PipelineHostResponse, skip_inheritance=False)
class PipelineHostResponse(RunspacePoolHostResponse):
    """PIPELINE_HOST_RESPONSE Message.

    Response from a host call executed on the client's host. Message is defined
    in `MS-PSRP 2.2.2.28 PIPELINE_HOST_RESPONSE`_.

    Args:
        ci: The Call ID associated with this operation.
        mi: The host method identifier.
        mr: The return value of the method.
        me: Exception thrown by a host method invocation.

    .. _MS-PSRP 2.2.2.28 PIPELINE_HOST_RESPONSE:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/d4298dce-ee0d-417d-a73a-b4ad26524e3b
    """


@PSMessageType(
    PSRPMessageType.ConnectRunspacePool,
    extended_properties=[
        PSNoteProperty("MinRunspaces", ps_type=PSInt),
        PSNoteProperty("MaxRunspaces", ps_type=PSInt),
    ],
)
class ConnectRunspacePool(PSObject):
    """CONNECT_RUNSPACEPOOL Message.

    Connect to a RunspacePool. Message is defined in
    `MS-PSRP 2.2.2.29 CONNECT_RUNSPACEPOOL`_.

    Note:
        This message is only used in ProtocolVersion>=2.2 (PowerShell v3.0+).

    Args:
        MinRunspaces: Minimum number of runspaces in the Runspace Pool.
        MaxRunspaces: Maximum number of runspaces in the RUnspace Pool.

    .. _MS-PSRP 2.2.2.29 CONNECT_RUNSPACEPOOL:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/9192146c-81b5-4abd-9b20-a56df272b95e
    """

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "ConnectRunspacePool",
        **kwargs: typing.Any,
    ) -> PSObject:
        use_string = True
        obj = PSObject()

        if instance.MinRunspaces:
            add_note_property(obj, "MinRunspaces", instance.MinRunspaces, PSInt)
            use_string = False

        if instance.MaxRunspaces:
            add_note_property(obj, "MaxRunspaces", instance.MaxRunspaces, PSInt)
            use_string = False

        # This is a weird object, will be '<S />' if neither count is specified.
        return PSString("") if use_string else obj


@PSMessageType(
    PSRPMessageType.RunspacePoolInitData,
    extended_properties=[
        PSNoteProperty("MinRunspaces", mandatory=True, ps_type=PSInt),
        PSNoteProperty("MaxRunspaces", mandatory=True, ps_type=PSInt),
    ],
)
class RunspacePoolInitData(PSObject):
    """RUNSPACEPOOL_INIT_DATA Message.

    RunspacePool initialization data. Message is defined in
    `MS-PSRP 2.2.2.30 RUNSPACEPOOL_INIT_DATA`_.

    Note:
        This message is only used in ProtocolVersion>=2.2 (PowerShell v3.0+).

    Args:
        MinRunspaces: Minimum number of runspaces in the Runspace Pool.
        MaxRunspaces: Maximum number of runspaces in the RUnspace Pool.

    .. _MS-PSRP 2.2.2.30 RUNSPACEPOOL_INIT_DATA:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/ee0ce0cb-2523-4d43-b8e8-049bb89112ad
    """


@PSMessageType(
    PSRPMessageType.ResetRunspaceState,
    extended_properties=[
        PSNoteProperty("ci", mandatory=True, ps_type=PSInt64),
    ],
)
class ResetRunspaceState(PSObject):
    """RESET_RUNSPACE_STATE Message.

    Reset RunspacePool Runspace state. Message is defined in
    `MS-PSRP 2.2.2.31 RESET_RUNSPACE_STATE`_.

    Note:
        This message is only used in ProtocolVersion>=2.3 (PowerShell v5.1+).

    Args:
        ci: The Call ID associated with this operation.

    .. _MS-PSRP 2.2.2.31 RESET_RUNSPACE_STATE:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/dc353f4b-c2e1-4172-a6ea-f72d7ef7c6bd
    """



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import datetime
import decimal
import re
import sys
import uuid
import xml.etree.ElementTree as ElementTree

import pytest

import psrpcore.types._primitive as primitive
from psrpcore.types import PSNoteProperty, PSObject

from ..conftest import (
    COMPLEX_ENCODED_STRING,
    COMPLEX_STRING,
    FakeCryptoProvider,
    deserialize,
    serialize,
)


def test_fail_to_create_integer_base():
    expected = re.escape(
        "Type PSIntegerBase cannot be instantiated; it can be used only as a base class for integer types."
    )

    with pytest.raises(TypeError, match=expected):
        primitive.PSIntegerBase(1)


@pytest.mark.parametrize(
    "ps_type, tag, type_names",
    [
        (primitive.PSString, "S", ["System.String", "System.Object"]),
        (primitive.PSUri, "URI", ["System.Uri", "System.Object"]),
        (primitive.PSXml, "XD", ["System.Xml.XmlDocument", "System.Xml.XmlNode", "System.Object"]),
        (primitive.PSScriptBlock, "SBK", ["System.Management.Automation.ScriptBlock", "System.Object"]),
    ],
)
def test_ps_string_types(ps_type, tag, type_names):
    ps_value = ps_type(COMPLEX_STRING)
    element = serialize(ps_value)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == f"<{tag}>{COMPLEX_ENCODED_STRING}</{tag}>"

    actual = deserialize(element)
    assert isinstance(actual, ps_type)
    assert isinstance(actual, str)
    assert actual == ps_value
    assert actual.PSObject.type_names == type_names

    # Check that we can still slice a string
    sliced_actual = actual[:6]
    assert isinstance(sliced_actual, ps_type)
    assert isinstance(sliced_actual, str)
    assert sliced_actual == COMPLEX_STRING[:6]


def test_ps_string_from_string():
    element = serialize(COMPLEX_STRING)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == f"<S>{COMPLEX_ENCODED_STRING}</S>"


@pytest.mark.parametrize(
    "ps_type, tag, type_names",
    [
        (primitive.PSString, "S", ["System.String", "System.Object"]),
        (primitive.PSUri, "URI", ["System.Uri", "System.Object"]),
        (primitive.PSXml, "XD", ["System.Xml.XmlDocument", "System.Xml.XmlNode", "System.Object"]),
        (primitive.PSScriptBlock, "SBK", ["System.Management.Automation.ScriptBlock", "System.Object"]),
    ],
)
def test_ps_string_with_properties(ps_type, tag, type_names):
    ps_value = ps_type(COMPLEX_STRING)
    ps_value.PSObject.extended_properties.append(PSNoteProperty("TestProperty"))
    ps_value.PSObject.extended_properties.append(PSNoteProperty(COMPLEX_STRING))
    ps_value.TestProperty = ps_type("property value")
    ps_value[COMPLEX_STRING] = ps_type("other value")
    element = serialize(ps_value)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == f'<Obj RefId="0"><{tag}>{COMPLEX_ENCODED_STRING}</{tag}>'
        f'<MS><{tag} N="TestProperty">property value</{tag}>'
        f'<{tag} N="{COMPLEX_ENCODED_STRING}">other value</{tag}>'
        f"</MS></Obj>"
    )

    actual = deserialize(element)
    assert isinstance(actual, ps_type)
    assert isinstance(actual, str)
    assert actual == ps_value

    assert actual["TestProperty"] == ps_type("property value")
    assert actual.TestProperty == ps_type("property value")
    assert isinstance(actual["TestProperty"], ps_type)

    assert actual[COMPLEX_STRING] == ps_type("other value")
    assert isinstance(actual[COMPLEX_STRING], ps_type)
    assert actual.PSObject.type_names == type_names

    # Check that we can still slice a string and the type is preserved
    sliced_actual = actual[:6]
    assert isinstance(sliced_actual, ps_type)
    assert isinstance(sliced_actual, str)
    assert sliced_actual == COMPLEX_STRING[:6]
    assert sliced_actual.PSObject.extended_properties == []

    # Check that a new instance does not inherit the same PSObject values
    new_str = ps_type("other")
    assert new_str.PSObject.adapted_properties == []
    assert new_str.PSObject.extended_properties == []


@pytest.mark.parametrize(
    "input_val",
    [
        "decimal",
        "text",
    ],
)
def test_ps_char(input_val):
    sparkles = b"\x28\x27".decode("utf-16-le")
    if input_val == "decimal":
        input_val = ord(sparkles)

    else:
        input_val = sparkles

    ps_char = primitive.PSChar(input_val)
    assert isinstance(ps_char, primitive.PSChar)
    assert isinstance(ps_char, int)
    assert str(ps_char) == sparkles

    element = serialize(ps_char)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == "<C>%s</C>" % int(ps_char)

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSChar)
    assert isinstance(actual, int)
    assert actual == ps_char
    assert int(actual) == ord(sparkles)
    assert str(actual) == sparkles
    assert actual.PSTypeNames == ["System.Char", "System.ValueType", "System.Object"]


@pytest.mark.parametrize(
    "input_val, expected",
    [
        (0, 0),
        ("\u0000", 0),
        (1, 1),
        ("1", 49),
        (b"\xc3\xa9", 233),
        ("é", 233),
        (65535, 65535),
        ("\uffff", 65535),
    ],
)
def test_ps_char_edge_cases(input_val, expected):
    str_expected = str(chr(expected))

    actual = primitive.PSChar(input_val)
    assert isinstance(actual, primitive.PSChar)
    assert actual == expected
    assert str(actual) == str_expected

    element = serialize(actual)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == "<C>%s</C>" % expected

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSChar)
    assert actual == expected
    assert str(actual) == str_expected
    assert actual.PSTypeNames == ["System.Char", "System.ValueType", "System.Object"]


@pytest.mark.parametrize(
    "input_val",
    [
        b"\xF0\x9D\x84\x9E".decode("utf-8"),
        "2c",
    ],
)
def test_ps_char_invalid_string(input_val):
    with pytest.raises(ValueError, match="A PSChar must be 1 UTF-16 codepoint"):
        primitive.PSChar(input_val)


@pytest.mark.parametrize("input_val", [-1, 65536])
def test_ps_char_invalid_int(input_val):
    with pytest.raises(ValueError, match="A PSChar must be between 0 and 65535."):
        primitive.PSChar(input_val)


def test_ps_char_with_properties():
    ps_char = primitive.PSChar("c")
    ps_char.PSObject.extended_properties.append(PSNoteProperty("Test Property"))
    ps_char["Test Property"] = 1

    element = serialize(ps_char)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == '<Obj RefId="0"><C>99</C><MS><I32 N="Test Property">1</I32></MS></Obj>'

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSChar)
    assert isinstance(actual, int)
    assert actual == 99
    assert str(actual) == "c"
    assert actual["Test Property"] == 1
    assert isinstance(actual["Test Property"], primitive.PSInt)
    assert actual.PSTypeNames == ["System.Char", "System.ValueType", "System.Object"]


@pytest.mark.parametrize(
    "input_val, expected",
    [
        (True, True),
        ("value", True),
        (1, True),
        (False, False),
        ("", False),
        (0, False),
    ],
)
def test_ps_bool(input_val, expected):
    actual = primitive.PSBool(input_val)
    assert isinstance(actual, primitive.PSBool)
    assert isinstance(actual, bool)
    assert actual == expected

    element = serialize(actual)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == "<B>%s</B>" % str(expected).lower()

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSBool)
    assert isinstance(actual, bool)
    assert not isinstance(actual, PSObject)  # We cannot subclass bool so this won't be a PSObject
    assert actual == expected


def test_ps_bool_from_bool():
    element = serialize(True)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == "<B>true</B>"


def test_ps_bool_deserialize_extended():
    # This just makes sure we don't choke on an extended primitive bool and we still get the raw value back.
    xml_val = '<Obj RefId="0"><B>true</B><MS><I32 N="Test Property">1</I32></MS></Obj>'
    element = ElementTree.fromstring(xml_val)

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSBool)
    assert isinstance(actual, bool)
    assert not isinstance(actual, PSObject)
    assert actual is True


@pytest.mark.parametrize(
    "input_val, expected, expected_str, expected_repr",
    [
        (
            datetime.datetime(1970, 1, 1, 0, 0, 0),
            "<DT>1970-01-01T00:00:00Z</DT>",
            "1970-01-01 00:00:00+00:00",
            "PSDateTime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc, nanosecond=0)",
        ),
        (
            datetime.datetime(1970, 1, 1, 0, 0, 0, microsecond=999999),
            "<DT>1970-01-01T00:00:00.999999Z</DT>",
            "1970-01-01 00:00:00.999999+00:00",
            "PSDateTime(1970, 1, 1, 0, 0, 0, 999999, tzinfo=datetime.timezone.utc, nanosecond=0)",
        ),
        (
            datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=datetime.timezone(offset=datetime.timedelta(hours=10))),
            "<DT>1970-01-01T00:00:00+10:00</DT>",
            "1970-01-01 00:00:00+10:00",
            "PSDateTime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone(datetime.timedelta(seconds=36000)), nanosecond=0)",
        ),
        (
            datetime.datetime(
                1970, 1, 1, 0, 0, 0, microsecond=999999, tzinfo=datetime.timezone(offset=datetime.timedelta(hours=10))
            ),
            "<DT>1970-01-01T00:00:00.999999+10:00</DT>",
            "1970-01-01 00:00:00.999999+10:00",
            "PSDateTime(1970, 1, 1, 0, 0, 0, 999999, tzinfo=datetime.timezone(datetime.timedelta(seconds=36000)), "
            "nanosecond=0)",
        ),
        (
            datetime.datetime(1600, 12, 12, 23, 59, 59),
            "<DT>1600-12-12T23:59:59Z</DT>",
            "1600-12-12 23:59:59+00:00",
            "PSDateTime(1600, 12, 12, 23, 59, 59, tzinfo=datetime.timezone.utc, nanosecond=0)",
        ),
    ],
)
def test_ps_datetime(input_val, expected, expected_str, expected_repr):
    # Python 3.6 has a slightly different repr for a timedelta value.
    if sys.version_info[:2] == (3, 6):
        expected_repr = expected_repr.replace("datetime.timedelta(seconds=", "datetime.timedelta(0, ")

    ps_datetime = primitive.PSDateTime(input_val)
    assert isinstance(ps_datetime, primitive.PSDateTime)
    assert isinstance(ps_datetime, datetime.datetime)

    element = serialize(ps_datetime)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == expected

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSDateTime)
    assert isinstance(actual, datetime.datetime)
    assert actual.year == input_val.year
    assert actual.month == input_val.month
    assert actual.day == input_val.day
    assert actual.hour == input_val.hour
    assert actual.minute == input_val.minute
    assert actual.second == input_val.second
    assert actual.microsecond == input_val.microsecond
    assert str(actual) == expected_str
    assert repr(actual) == expected_repr
    assert actual.PSTypeNames == ["System.DateTime", "System.ValueType", "System.Object"]


def test_ps_datetime_from_datetime():
    element = serialize(datetime.datetime(1970, 1, 1, 0, 0, 0, microsecond=999999))
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()

    assert actual == "<DT>1970-01-01T00:00:00.999999Z</DT>"


@pytest.mark.parametrize("nanosecond, fraction", [(7, 0), (70, 0), (700, 7)])
def test_ps_datetime_nanosecond(nanosecond, fraction):
    ps_datetime = primitive.PSDateTime(1970, 6, 11, 4, 8, 23, microsecond=123456, nanosecond=nanosecond)
    assert isinstance(ps_datetime, primitive.PSDateTime)
    assert isinstance(ps_datetime, datetime.datetime)

    element = serialize(ps_datetime)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == "<DT>1970-06-11T04:08:23.123456%sZ</DT>" % fraction

    assert str(ps_datetime) == f"1970-06-11 04:08:23.123456{nanosecond:03d}"


def test_ps_datetime_nanosecond_timezone():
    tz = datetime.timezone(offset=datetime.timedelta(hours=-10, minutes=-35))
    ps_datetime = primitive.PSDateTime(1970, 6, 11, 4, 8, 23, microsecond=123456, nanosecond=454, tzinfo=tz)
    assert isinstance(ps_datetime, primitive.PSDateTime)
    assert isinstance(ps_datetime, datetime.datetime)

    element = serialize(ps_datetime)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == "<DT>1970-06-11T04:08:23.1234564-10:35</DT>"

    assert str(ps_datetime) == "1970-06-11 04:08:23.123456454-10:35"


def test_ps_datetime_with_properties():
    ps_datetime = primitive.PSDateTime(2000, 2, 29, 15, 43, 10, microsecond=10)
    ps_datetime.PSObject.extended_properties.append(PSNoteProperty("Test Property"))
    ps_datetime["Test Property"] = 1

    element = serialize(ps_datetime)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0"><DT>2000-02-29T15:43:10.000010Z</DT>' '<MS><I32 N="Test Property">1</I32></MS></Obj>'
    )

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSDateTime)
    assert isinstance(actual, datetime.datetime)
    assert actual.year == ps_datetime.year
    assert actual.month == ps_datetime.month
    assert actual.day == ps_datetime.day
    assert actual.hour == ps_datetime.hour
    assert actual.minute == ps_datetime.minute
    assert actual.second == ps_datetime.second
    assert actual.microsecond == ps_datetime.microsecond
    assert str(actual) == "2000-02-29 15:43:10.000010+00:00"
    assert actual["Test Property"] == 1
    assert isinstance(actual["Test Property"], primitive.PSInt)
    assert actual.PSTypeNames == ["System.DateTime", "System.ValueType", "System.Object"]


def test_ps_datetime_add_duration():
    datetime_obj = primitive.PSDateTime(2000, 2, 29, 15, 43, 10, microsecond=10, nanosecond=733)
    duration = primitive.PSDuration(days=400, hours=10, minutes=20, seconds=55, microseconds=100, nanoseconds=400)

    actual = datetime_obj + duration

    assert isinstance(actual, primitive.PSDateTime)
    assert isinstance(actual, datetime.datetime)
    assert actual.year == 2001
    assert actual.month == 4
    assert actual.day == 5
    assert actual.hour == 2
    assert actual.minute == 4
    assert actual.second == 5
    assert actual.microsecond == 111
    assert actual.nanosecond == 133

    element = serialize(actual)
    actual_str = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual_str == "<DT>2001-04-05T02:04:05.0001111Z</DT>"


def test_ps_datetime_sub_duration():
    datetime_obj = primitive.PSDateTime(2000, 2, 29, 15, 43, 10, microsecond=10, nanosecond=733)
    duration = primitive.PSDuration(days=400, hours=10, minutes=20, seconds=55, microseconds=100, nanoseconds=400)

    actual = datetime_obj - duration

    assert isinstance(actual, primitive.PSDateTime)
    assert isinstance(actual, datetime.datetime)
    assert actual.year == 1999
    assert actual.month == 1
    assert actual.day == 25
    assert actual.hour == 5
    assert actual.minute == 22
    assert actual.second == 14
    assert actual.microsecond == 999910
    assert actual.nanosecond == 333

    element = serialize(actual)
    actual_str = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual_str == "<DT>1999-01-25T05:22:14.9999103Z</DT>"


def test_ps_datetime_sub_datetime():
    datetime_obj = primitive.PSDateTime(2000, 2, 29, 15, 43, 10, microsecond=10, nanosecond=731)
    sub_datetime = primitive.PSDateTime(1970, 1, 1, nanosecond=732)

    actual = datetime_obj - sub_datetime

    assert isinstance(actual, primitive.PSDuration)
    assert isinstance(actual, datetime.timedelta)
    assert actual.days == 11016
    assert actual.seconds == 56590
    assert actual.microseconds == 9
    assert actual.nanoseconds == 999


@pytest.mark.parametrize(
    "input_val, expected, expected_str, expected_repr",
    [
        (datetime.timedelta(0), "<TS>PT0S</TS>", "0:00:00", "PSDuration(0)"),
        (datetime.timedelta(hours=24), "<TS>P1D</TS>", "1 day, 0:00:00", "PSDuration(days=1)"),
        (datetime.timedelta(hours=1, minutes=5), "<TS>PT1H5M</TS>", "1:05:00", "PSDuration(seconds=3900)"),
        (datetime.timedelta(seconds=6005), "<TS>PT1H40M5S</TS>", "1:40:05", "PSDuration(seconds=6005)"),
        (
            datetime.timedelta(seconds=-6005),
            "<TS>-PT1H40M5S</TS>",
            "-1 day, 22:19:55",
            "PSDuration(days=-1, seconds=80395)",
        ),
        (datetime.timedelta(microseconds=99), "<TS>PT0.000099S</TS>", "0:00:00.000099", "PSDuration(microseconds=99)"),
        (
            datetime.timedelta(microseconds=922337203685477580),
            "<TS>P10675199DT2H48M5.47758S</TS>",
            "10675199 days, 2:48:05.477580",
            "PSDuration(days=10675199, seconds=10085, microseconds=477580)",
        ),
    ],
)
def test_ps_duration(input_val, expected, expected_str, expected_repr):
    ps_duration = primitive.PSDuration(input_val)
    assert isinstance(ps_duration, primitive.PSDuration)
    assert isinstance(ps_duration, datetime.timedelta)

    element = serialize(ps_duration)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == expected

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSDuration)
    assert isinstance(actual, datetime.timedelta)
    assert actual.days == input_val.days
    assert actual.microseconds == input_val.microseconds
    assert actual.seconds == input_val.seconds
    assert str(actual) == expected_str
    assert repr(actual) == expected_repr
    assert actual.PSTypeNames == ["System.TimeSpan", "System.ValueType", "System.Object"]


def test_ps_duration_from_timedelta():
    element = serialize(datetime.timedelta(microseconds=922337203685477580))

    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == "<TS>P10675199DT2H48M5.47758S</TS>"


@pytest.mark.parametrize("nanosecond", [8, 80, 800])
def test_ps_duration_with_nanoseconds(nanosecond):
    base_nanoseconds = 922337203685477580000
    fraction = ""
    if nanosecond > 100:
        fraction = f"0{nanosecond // 100}"

    ps_duration = primitive.PSDuration(nanoseconds=(base_nanoseconds + nanosecond))
    assert isinstance(ps_duration, primitive.PSDuration)
    assert isinstance(ps_duration, datetime.timedelta)

    element = serialize(ps_duration)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == f"<TS>P10675199DT2H48M5.47758{fraction}S</TS>"

    assert (
        repr(ps_duration) == f"PSDuration(days=10675199, seconds=10085, microseconds=477580, "
        f"nanoseconds={nanosecond})"
    )
    assert str(ps_duration) == f"10675199 days, 2:48:05.477580{nanosecond:03d}"


def test_ps_duration_with_properties():
    ps_duration = primitive.PSDuration(days=10, hours=25, minutes=70, seconds=129, microseconds=1000, nanoseconds=1100)
    ps_duration.PSObject.extended_properties.append(PSNoteProperty("Test Property"))
    ps_duration["Test Property"] = 1

    element = serialize(ps_duration)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == '<Obj RefId="0"><TS>P11DT2H12M9.0010011S</TS><MS><I32 N="Test Property">1</I32></MS></Obj>'

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSDuration)
    assert isinstance(actual, datetime.timedelta)
    assert actual.days == ps_duration.days
    assert actual.microseconds == ps_duration.microseconds
    assert actual.seconds == ps_duration.seconds
    assert actual.microseconds == ps_duration.microseconds
    assert actual.nanoseconds == ps_duration.nanoseconds
    assert str(actual) == "11 days, 2:12:09.001001100"
    assert actual["Test Property"] == 1
    assert isinstance(actual["Test Property"], primitive.PSInt)
    assert actual.PSTypeNames == ["System.TimeSpan", "System.ValueType", "System.Object"]


@pytest.mark.parametrize(
    "other, expected",
    [
        (
            primitive.PSDuration(days=30, hours=23, minutes=50, seconds=45, microseconds=19, nanoseconds=521),
            primitive.PSDuration(days=41, hours=10, minutes=4, seconds=44, microseconds=2013, nanoseconds=175),
        ),
        (
            primitive.PSDuration(days=-30, hours=23, minutes=-50, seconds=45, microseconds=-19, nanoseconds=521),
            primitive.PSDuration(days=-19, hours=8, minutes=24, seconds=44, microseconds=1975, nanoseconds=175),
        ),
        (
            datetime.timedelta(days=30, hours=23, minutes=50, seconds=45, microseconds=19),
            primitive.PSDuration(days=41, hours=10, minutes=4, seconds=44, microseconds=2012, nanoseconds=654),
        ),
        (
            datetime.timedelta(days=-30, hours=23, minutes=-50, seconds=45, microseconds=-19),
            primitive.PSDuration(days=-19, hours=8, minutes=24, seconds=44, microseconds=1974, nanoseconds=654),
        ),
    ],
)
def test_ps_duration_add(other, expected):
    duration = primitive.PSDuration(days=10, hours=10, minutes=13, seconds=59, microseconds=1993, nanoseconds=654)

    actual = duration + other
    assert actual == expected


def test_ps_duration_add_invalid_type():
    duration = primitive.PSDuration(1)

    with pytest.raises(TypeError):
        duration += 1


@pytest.mark.parametrize(
    "other, expected",
    [
        (
            primitive.PSDuration(days=30, hours=23, minutes=50, seconds=45, microseconds=19, nanoseconds=521),
            primitive.PSDuration(days=-21, hours=10, minutes=23, seconds=14, microseconds=1974, nanoseconds=133),
        ),
        (
            primitive.PSDuration(days=-30, hours=23, minutes=-50, seconds=45, microseconds=-19, nanoseconds=521),
            primitive.PSDuration(days=39, hours=12, minutes=3, seconds=14, microseconds=2012, nanoseconds=133),
        ),
        (
            datetime.timedelta(days=30, hours=23, minutes=50, seconds=45, microseconds=19),
            primitive.PSDuration(days=-21, hours=10, minutes=23, seconds=14, microseconds=1974, nanoseconds=654),
        ),
        (
            datetime.timedelta(days=-30, hours=23, minutes=-50, seconds=45, microseconds=-19),
            primitive.PSDuration(days=39, hours=12, minutes=3, seconds=14, microseconds=2012, nanoseconds=654),
        ),
    ],
)
def test_ps_duration_sub(other, expected):
    duration = primitive.PSDuration(days=10, hours=10, minutes=13, seconds=59, microseconds=1993, nanoseconds=654)

    actual = duration - other
    assert actual == expected


def test_ps_duration_sub_invalid_type():
    duration = primitive.PSDuration(1)

    with pytest.raises(TypeError):
        duration -= 1


@pytest.mark.parametrize(
    "other, expected",
    [
        (
            datetime.timedelta(days=30, hours=23, minutes=50, seconds=45, microseconds=19),
            primitive.PSDuration(days=20, hours=13, minutes=36, seconds=45, microseconds=998025, nanoseconds=346),
        ),
        (
            datetime.timedelta(days=-30, hours=23, minutes=-50, seconds=45, microseconds=-19),
            primitive.PSDuration(days=-40, hours=11, minutes=56, seconds=45, microseconds=997987, nanoseconds=346),
        ),
    ],
)
def test_ps_duration_rsub(other, expected):
    duration = primitive.PSDuration(days=10, hours=10, minutes=13, seconds=59, microseconds=1993, nanoseconds=654)

    actual = other - duration
    assert actual == expected


def test_ps_duration_negative():
    original = primitive.PSDuration(days=30, hours=10, nanoseconds=500)

    pos = +original
    assert isinstance(pos, primitive.PSDuration)
    assert original == +original

    sub = -original
    assert isinstance(sub, primitive.PSDuration)
    assert sub.days == -31
    assert sub.seconds == 50399
    assert sub.microseconds == 999999
    assert sub.nanoseconds == 500

    pos = -sub
    assert pos == original


def test_duration_equality():
    duration_lowest = primitive.PSDuration(nanoseconds=1)
    duration_lower = primitive.PSDuration(nanoseconds=2)
    duration_lower2 = primitive.PSDuration(nanoseconds=2)
    duration_higher = primitive.PSDuration(nanoseconds=3)
    duration_higher2 = primitive.PSDuration(nanoseconds=3)
    duration_highest = primitive.PSDuration(nanoseconds=4)

    assert duration_lower == duration_lower2
    assert duration_lowest != duration_lower
    assert duration_lowest < duration_lower
    assert duration_lowest <= duration_lower
    assert not duration_lower < duration_lower2
    assert duration_lower <= duration_lower2
    assert duration_lower < duration_higher
    assert duration_highest > duration_higher
    assert duration_highest >= duration_higher
    assert duration_higher >= duration_higher
    assert not duration_higher > duration_higher2
    assert duration_higher >= duration_higher2


def test_duration_timedelta_equality():
    duration_low = primitive.PSDuration(days=1)
    duration_low_ns = primitive.PSDuration(days=1, nanoseconds=1)
    timedelta_low = datetime.timedelta(days=1)

    duration_high = primitive.PSDuration(days=2, nanoseconds=1)
    timedelta_high = datetime.timedelta(days=2)

    assert duration_low == timedelta_low
    assert timedelta_low == duration_low
    assert not duration_low_ns == timedelta_low
    assert not timedelta_low == duration_low_ns

    assert duration_low < timedelta_high
    assert duration_low_ns < timedelta_high
    assert duration_low <= timedelta_high
    assert duration_low_ns <= timedelta_high

    assert timedelta_high > duration_low_ns
    assert timedelta_high >= duration_low_ns

    assert duration_high > timedelta_low
    assert duration_high >= timedelta_low


@pytest.mark.parametrize(
    "ps_type, tag, type_name",
    [
        (primitive.PSByte, "By", "Byte"),
        (primitive.PSSByte, "SB", "SByte"),
        (primitive.PSUInt16, "U16", "UInt16"),
        (primitive.PSInt16, "I16", "Int16"),
        (primitive.PSUInt, "U32", "UInt32"),
        (primitive.PSInt, "I32", "Int32"),
        (primitive.PSUInt64, "U64", "UInt64"),
        (primitive.PSInt64, "I64", "Int64"),
    ],
)
def test_numeric(ps_type, tag, type_name):
    for value in [ps_type.MinValue, 0, 1, ps_type.MaxValue]:
        ps_value = ps_type(value)
        assert isinstance(ps_value, ps_type)
        assert isinstance(ps_value, int)

        element = serialize(ps_value)
        actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
        assert actual == f"<{tag}>{value}</{tag}>"

        actual = deserialize(element)
        assert isinstance(actual, ps_type)
        assert isinstance(actual, int)
        assert actual == value
        assert actual.PSTypeNames == [f"System.{type_name}", "System.ValueType", "System.Object"]


@pytest.mark.parametrize(
    "input_value, tag",
    [
        (10, "I32"),
        (256, "I32"),
        (2147483647, "I32"),
        (2147483648, "I64"),
    ],
)
def test_ps_int_from_int(input_value, tag):
    element = serialize(input_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == f"<{tag}>{input_value}</{tag}>"


@pytest.mark.parametrize(
    "input_args",
    [
        ("11111111", 2),
        ("377", 8),
        ("FF", 16),
    ],
)
def test_ps_int_with_base(input_args):
    ps_value = primitive.PSInt(*input_args)
    assert isinstance(ps_value, primitive.PSInt)
    assert isinstance(ps_value, int)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == f"<I32>255</I32>"


@pytest.mark.parametrize(
    "ps_type, tag, type_name",
    [
        (primitive.PSByte, "By", "Byte"),
        (primitive.PSSByte, "SB", "SByte"),
        (primitive.PSUInt16, "U16", "UInt16"),
        (primitive.PSInt16, "I16", "Int16"),
        (primitive.PSUInt, "U32", "UInt32"),
        (primitive.PSInt, "I32", "Int32"),
        (primitive.PSUInt64, "U64", "UInt64"),
        (primitive.PSInt64, "I64", "Int64"),
    ],
)
def test_numeric_with_properties(ps_type, tag, type_name):
    ps_value = ps_type(None)
    assert isinstance(ps_value, ps_type)
    assert ps_value == 0

    ps_value = ps_type(10)
    ps_value.PSObject.extended_properties.append(PSNoteProperty("Test Property"))
    ps_value["Test Property"] = 1

    element = serialize(ps_value)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == f'<Obj RefId="0"><{tag}>10</{tag}><MS><I32 N="Test Property">1</I32></MS></Obj>'

    actual = deserialize(element)
    assert isinstance(actual, ps_type)
    assert isinstance(actual, int)
    assert actual == 10
    assert actual["Test Property"] == 1
    assert isinstance(actual["Test Property"], primitive.PSInt)
    assert actual.PSTypeNames == [f"System.{type_name}", "System.ValueType", "System.Object"]


@pytest.mark.parametrize(
    "ps_type, value",
    [
        (primitive.PSByte, -1),
        (primitive.PSByte, 256),
        (primitive.PSSByte, -129),
        (primitive.PSSByte, 128),
        (primitive.PSUInt16, -1),
        (primitive.PSUInt16, 65536),
        (primitive.PSInt16, -32769),
        (primitive.PSInt16, 32768),
        (primitive.PSUInt, -1),
        (primitive.PSUInt, 4294967296),
        (primitive.PSInt, -2147483649),
        (primitive.PSInt, 2147483648),
        (primitive.PSUInt64, -1),
        (primitive.PSUInt64, 18446744073709551616),
        (primitive.PSInt64, -9223372036854775809),
        (primitive.PSInt64, 9223372036854775808),
    ],
)
def test_numeric_invalid_value(ps_type, value):
    expected = re.escape(
        f"Cannot create {ps_type.__qualname__} with value '{value}': Value must be between "
        f"{ps_type.MinValue} and {ps_type.MaxValue}."
    )
    with pytest.raises(ValueError, match=expected):
        ps_type(value)


@pytest.mark.parametrize(
    "ps_type",
    [
        primitive.PSByte,
        primitive.PSSByte,
        primitive.PSUInt16,
        primitive.PSInt16,
        primitive.PSUInt,
        primitive.PSInt,
        primitive.PSUInt64,
        primitive.PSInt64,
    ],
)
def test_numeric_operators(ps_type):
    actual = ps_type(2) + 4
    assert isinstance(actual, ps_type)
    assert actual == 6

    actual = ps_type(3) & 2
    assert isinstance(actual, ps_type)
    assert actual == 2

    quotient, remainder = divmod(ps_type(11), 10)
    assert isinstance(quotient, ps_type)
    assert quotient == 1
    assert isinstance(remainder, int)
    assert remainder == 1

    actual = ps_type(13) // 2
    assert isinstance(actual, ps_type)
    assert actual == 6

    actual = ps_type(1) << 2
    assert isinstance(actual, ps_type)
    assert actual == 4

    actual = ps_type(3) % 2
    assert isinstance(actual, ps_type)
    assert actual == 1

    actual = ps_type(3) * 2
    assert isinstance(actual, ps_type)
    assert actual == 6

    actual = ps_type(1) | 2
    assert isinstance(actual, ps_type)
    assert actual == 3

    actual = ps_type(2) ** 2
    assert isinstance(actual, ps_type)
    assert actual == 4

    actual = ps_type(4) >> 2
    assert isinstance(actual, ps_type)
    assert actual == 1

    actual = ps_type(4) - 2
    assert isinstance(actual, ps_type)
    assert actual == 2

    actual = ps_type(19) ^ 21
    assert isinstance(actual, ps_type)
    assert actual == 6

    actual = ps_type(1)
    actual += 1
    assert isinstance(actual, ps_type)
    assert actual == 2

    actual -= 1
    assert isinstance(actual, ps_type)
    assert actual == 1

    actual *= 2
    assert isinstance(actual, ps_type)
    assert actual == 2


@pytest.mark.parametrize(
    "ps_type",
    [
        primitive.PSSByte,
        primitive.PSInt16,
        primitive.PSInt,
        primitive.PSInt64,
    ],
)
def test_numeric_negative_operators(ps_type):
    actual = abs(ps_type(-1))
    assert isinstance(actual, ps_type)
    assert actual == 1

    actual = ~ps_type(-1)
    assert isinstance(actual, ps_type)
    assert actual == 0

    actual = ~ps_type(1)
    assert isinstance(actual, ps_type)
    assert actual == -2

    actual = -ps_type(-1)
    assert isinstance(actual, ps_type)
    assert actual == 1

    actual = -ps_type(1)
    assert isinstance(actual, ps_type)
    assert actual == -1

    actual = +ps_type(-1)
    assert isinstance(actual, ps_type)
    assert actual == -1

    actual = +ps_type(1)
    assert isinstance(actual, ps_type)
    assert actual == 1


@pytest.mark.parametrize(
    "ps_type, input_val, expected, type_name",
    [
        (primitive.PSSingle, 1, "<Sg>1.0</Sg>", "Single"),
        (primitive.PSSingle, 1.0, "<Sg>1.0</Sg>", "Single"),
        (primitive.PSSingle, 1.1, "<Sg>1.1</Sg>", "Single"),
        (primitive.PSSingle, 3.402823e38, "<Sg>3.402823E+38</Sg>", "Single"),
        (primitive.PSSingle, -3.402823e38, "<Sg>-3.402823E+38</Sg>", "Single"),
        (primitive.PSDouble, 1, "<Db>1.0</Db>", "Double"),
        (primitive.PSDouble, 1.0, "<Db>1.0</Db>", "Double"),
        (primitive.PSDouble, 1.1, "<Db>1.1</Db>", "Double"),
        (primitive.PSDouble, 1.7976931348623157e308, "<Db>1.7976931348623157E+308</Db>", "Double"),
        (primitive.PSDouble, 1.79769313486232e308, "<Db>INF</Db>", "Double"),
        (primitive.PSDouble, -1.79769313486232e307, "<Db>-1.79769313486232E+307</Db>", "Double"),
        (primitive.PSDouble, -1.79769313486232e308, "<Db>-INF</Db>", "Double"),
    ],
)
def test_ps_single_and_double(ps_type, input_val, expected, type_name):
    ps_value = ps_type(input_val)
    assert isinstance(ps_value, ps_type)
    assert isinstance(ps_value, float)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == expected

    actual = deserialize(element)
    assert isinstance(actual, ps_type)
    assert isinstance(actual, float)
    assert actual == float(input_val)
    assert actual.PSTypeNames == [f"System.{type_name}", "System.ValueType", "System.Object"]


def test_ps_single_from_float():
    element = serialize(1.1)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == "<Sg>1.1</Sg>"


@pytest.mark.parametrize(
    "ps_type, tag, type_name",
    [
        (primitive.PSSingle, "Sg", "Single"),
        (primitive.PSDouble, "Db", "Double"),
    ],
)
def test_ps_single_and_double_with_properties(ps_type, tag, type_name):
    ps_value = ps_type(1.1)
    ps_value.PSObject.extended_properties.append(PSNoteProperty("Test Property"))
    ps_value["Test Property"] = 1.1

    element = serialize(ps_value)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == f'<Obj RefId="0"><{tag}>1.1</{tag}><MS><Sg N="Test Property">1.1</Sg></MS></Obj>'

    actual = deserialize(element)
    assert isinstance(actual, ps_type)
    assert isinstance(actual, float)
    assert actual == 1.1
    assert actual["Test Property"] == 1.1
    assert isinstance(actual["Test Property"], primitive.PSSingle)
    assert actual.PSTypeNames == [f"System.{type_name}", "System.ValueType", "System.Object"]


@pytest.mark.parametrize(
    "input_value, expected",
    [
        (0, "<D>0</D>"),
        (-1, "<D>-1</D>"),
        (1, "<D>1</D>"),
        (1.0, "<D>1</D>"),
        ("1.1", "<D>1.1</D>"),
        ("1.10", "<D>1.10</D>"),
    ],
)
def test_ps_decimal(input_value, expected):
    ps_value = primitive.PSDecimal(input_value)
    assert isinstance(ps_value, primitive.PSDecimal)
    assert isinstance(ps_value, decimal.Decimal)
    assert str(ps_value) == decimal.Decimal.__str__(ps_value)
    assert repr(ps_value) == decimal.Decimal.__repr__(ps_value)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == expected

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSDecimal)
    assert isinstance(actual, decimal.Decimal)
    assert actual == decimal.Decimal(input_value)
    assert actual.PSTypeNames == ["System.Decimal", "System.ValueType", "System.Object"]


def test_ps_decimal_with_properties():
    ps_value = primitive.PSDecimal("1.302000")
    ps_value.PSObject.extended_properties.append(PSNoteProperty("Test Property"))
    ps_value["Test Property"] = decimal.Decimal(0)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == f'<Obj RefId="0"><D>1.302000</D><MS><D N="Test Property">0</D></MS></Obj>'

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSDecimal)
    assert isinstance(actual, decimal.Decimal)
    assert actual == decimal.Decimal("1.302000")
    assert actual["Test Property"] == decimal.Decimal(0)
    assert isinstance(actual["Test Property"], primitive.PSDecimal)
    assert actual.PSTypeNames == ["System.Decimal", "System.ValueType", "System.Object"]


@pytest.mark.parametrize(
    "input_value, expected",
    [
        (b"\x00\x01\x02\x03", "<BA>AAECAw==</BA>"),
        (b"", "<BA />"),
    ],
)
def test_ps_byte_array(input_value, expected):
    ps_value = primitive.PSByteArray(input_value)
    assert isinstance(ps_value, primitive.PSByteArray)
    assert isinstance(ps_value, bytes)
    assert str(ps_value) == bytes.__str__(ps_value)
    assert repr(ps_value) == bytes.__repr__(ps_value)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == expected

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSByteArray)
    assert isinstance(actual, bytes)
    assert actual == input_value
    assert actual.PSTypeNames == ["System.Byte[]", "System.Array", "System.Object"]


def test_ps_byte_array_with_properties():
    value = "café".encode("utf-8")

    ps_value = primitive.PSByteArray(value)
    ps_value.PSObject.extended_properties.append(PSNoteProperty("Test Property"))
    ps_value["Test Property"] = value

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == f'<Obj RefId="0"><BA>Y2Fmw6k=</BA><MS><BA N="Test Property">Y2Fmw6k=</BA></MS></Obj>'

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSByteArray)
    assert isinstance(actual, bytes)
    assert actual == value
    assert actual["Test Property"] == value
    assert isinstance(actual["Test Property"], primitive.PSByteArray)
    assert actual.PSTypeNames == ["System.Byte[]", "System.Array", "System.Object"]

    # Check that we can still slice bytes and the type is preserved
    sliced_actual = actual[:2]
    assert isinstance(sliced_actual, primitive.PSByteArray)
    assert isinstance(sliced_actual, bytes)
    assert sliced_actual == value[:2]
    assert sliced_actual.PSObject.extended_properties == []

    # Check that a new PSString instance does not inherit the same PSObject values
    new_str = primitive.PSByteArray(b"other")
    assert new_str.PSObject.adapted_properties == []
    assert new_str.PSObject.extended_properties == []


@pytest.mark.parametrize(
    "input_value, expected",
    [
        ("00000000-0000-0000-0000-000000000000", "<G>00000000-0000-0000-0000-000000000000</G>"),
        ("f5853fa8-a3d8-438c-bf94-723d3fef8934", "<G>f5853fa8-a3d8-438c-bf94-723d3fef8934</G>"),
        (uuid.UUID("f5853fa8-a3d8-438c-bf94-723d3fef8934"), "<G>f5853fa8-a3d8-438c-bf94-723d3fef8934</G>"),
    ],
)
def test_ps_guid(input_value, expected):
    ps_value = primitive.PSGuid(input_value)
    assert isinstance(ps_value, primitive.PSGuid)
    assert isinstance(ps_value, uuid.UUID)
    assert str(ps_value) == uuid.UUID.__str__(ps_value)
    assert repr(ps_value) == uuid.UUID.__repr__(ps_value)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == expected

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSGuid)
    assert isinstance(actual, uuid.UUID)
    assert actual == primitive.PSGuid(input_value)
    assert actual.PSTypeNames == ["System.Guid", "System.ValueType", "System.Object"]


def test_ps_guid_with_properties():
    ps_value = primitive.PSGuid(int=1)
    ps_value.PSObject.extended_properties.append(PSNoteProperty("Test Property"))
    ps_value["Test Property"] = uuid.UUID(int=2)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert (
        actual == f'<Obj RefId="0"><G>00000000-0000-0000-0000-000000000001</G>'
        f'<MS><G N="Test Property">00000000-0000-0000-0000-000000000002</G></MS></Obj>'
    )

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSGuid)
    assert isinstance(actual, uuid.UUID)
    assert actual == uuid.UUID(int=1)
    assert actual["Test Property"] == uuid.UUID(int=2)
    assert isinstance(actual["Test Property"], primitive.PSGuid)
    assert actual.PSTypeNames == ["System.Guid", "System.ValueType", "System.Object"]

    # uuid.UUID uses __slots__ so setting PSObject on the actual instance is a bit tricker. This makes sure we've done
    # it correctly.
    ps_value = primitive.PSGuid(int=0)
    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == "<G>00000000-0000-0000-0000-000000000000</G>"


@pytest.mark.parametrize("input_val", [primitive.PSNull, None])
def test_ps_null(input_val):
    element = serialize(input_val)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == "<Nil />"

    actual = deserialize(element)
    assert actual is None
    assert actual is primitive.PSNull


@pytest.mark.parametrize(
    "input_value",
    [
        "0",
        "01.1",
        "1.2.3.4.5",
        "1.2.3a",
        "1.2.3.a",
        "1.01",
        "1.0.",
        "1.0.0.",
        "1.0.0.0.",
    ],
)
def test_ps_version_invalid_strings(input_value):
    expected = re.escape(
        f"Invalid PSVersion string '{input_value}': must be 2 to 4 groups of numbers that are " f"separated by '.'"
    )

    with pytest.raises(ValueError, match=expected):
        primitive.PSVersion(input_value)


def test_ps_version_no_major_and_minor():
    expected = "The major and minor versions must be specified"

    with pytest.raises(ValueError, match=expected):
        primitive.PSVersion(major=1)

    with pytest.raises(ValueError, match=expected):
        primitive.PSVersion(minor=1)


def test_ps_version_build_not_set():
    with pytest.raises(ValueError, match="The build version must be set when revision is set"):
        primitive.PSVersion(major=1, minor=0, revision=0)


@pytest.mark.parametrize(
    "input_value, major, minor, build, revision",
    [
        ("1.0", 1, 0, None, None),
        ("1.1", 1, 1, None, None),
        ("1.2.3", 1, 2, 3, None),
        ("1.2.3.4", 1, 2, 3, 4),
        ("0.0", 0, 0, None, None),
        ("0.1", 0, 1, None, None),
        ("0.0.1", 0, 0, 1, None),
        ("1.0.1", 1, 0, 1, None),
        ("10.10234.2030.102", 10, 10234, 2030, 102),
    ],
)
def test_ps_version(input_value, major, minor, build, revision):
    ps_value = primitive.PSVersion(input_value)
    assert isinstance(ps_value, primitive.PSVersion)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == f"<Version>{input_value}</Version>"

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSVersion)
    assert actual == ps_value
    assert actual.major == major
    assert actual.minor == minor
    assert actual.build == build
    assert actual.revision == revision
    assert actual.PSTypeNames == ["System.Version", "System.Object"]


def test_ps_version_with_properties():
    ps_value = primitive.PSVersion("1.1")
    ps_value.PSObject.extended_properties.append(PSNoteProperty("Test Property"))
    ps_value["Test Property"] = primitive.PSVersion("0.0")

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == f'<Obj RefId="0"><Version>1.1</Version>' f'<MS><Version N="Test Property">0.0</Version></MS></Obj>'

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSVersion)
    assert actual == ps_value
    assert actual.major == 1
    assert actual.minor == 1
    assert actual.build is None
    assert actual.revision is None
    assert actual["Test Property"] == "0.0"
    assert isinstance(actual["Test Property"], primitive.PSVersion)
    assert actual.PSTypeNames == ["System.Version", "System.Object"]


@pytest.mark.parametrize(
    "input_value, expected_repr",
    [
        ("1.0", "major=1, minor=0"),
        ("0.0.1", "major=0, minor=0, build=1"),
        ("1.0.1", "major=1, minor=0, build=1"),
        ("1.0.1.0", "major=1, minor=0, build=1, revision=0"),
        ("1.0.1.99", "major=1, minor=0, build=1, revision=99"),
    ],
)
def test_ps_version_str(input_value, expected_repr):
    actual = primitive.PSVersion(input_value)
    assert str(actual) == input_value
    assert repr(actual) == f"PSVersion({expected_repr})"


@pytest.mark.parametrize(
    "version, other, expected",
    [
        (primitive.PSVersion("1.0"), primitive.PSVersion("1.0"), True),
        (primitive.PSVersion("1.0"), "1.0", True),
        (primitive.PSVersion("1.1"), primitive.PSVersion("1.0"), False),
        (primitive.PSVersion("1.1"), "1.0", False),
        (primitive.PSVersion("1.0.0"), primitive.PSVersion("1.0"), False),
    ],
)
def test_ps_version_equals(version, other, expected):
    assert (version == other) == expected


@pytest.mark.parametrize(
    "version, other, expected",
    [
        ("2.2", "2.3", False),
        ("2.3", "2.3", False),
        ("2.4", "2.3", True),
        ("3.0", "2.3", True),
        ("1.0", "2.3", False),
        ("2.3.0", "2.3", True),
        ("2.3.1", "2.3", True),
        ("2.3", "2.3.0", False),
        ("2.3", "2.3.1", False),
        ("99.102.0.19", "99.102.1.0", False),
        ("99.102.0.19", "99.102.0.0", True),
        ("99.102.0.19", "99.102.0.19", False),
    ],
)
def test_ps_version_greater_than(version, other, expected):
    assert (primitive.PSVersion(version) > other) == expected


@pytest.mark.parametrize(
    "version, other, expected",
    [
        ("2.2", "2.3", False),
        ("2.3", "2.3", True),
        ("2.4", "2.3", True),
        ("3.0", "2.3", True),
        ("1.0", "2.3", False),
        ("2.3.0", "2.3", True),
        ("2.3.1", "2.3", True),
        ("2.3", "2.3.0", False),
        ("2.3", "2.3.1", False),
        ("99.102.0.19", "99.102.1.0", False),
        ("99.102.0.19", "99.102.0.0", True),
        ("99.102.0.19", "99.102.0.19", True),
    ],
)
def test_ps_version_greater_or_equal(version, other, expected):
    assert (primitive.PSVersion(version) >= other) == expected


@pytest.mark.parametrize(
    "version, other, expected",
    [
        ("2.2", "2.3", True),
        ("2.3", "2.3", False),
        ("2.4", "2.3", False),
        ("3.0", "2.3", False),
        ("1.0", "2.3", True),
        ("2.3.0", "2.3", False),
        ("2.3.1", "2.3", False),
        ("2.3", "2.3.0", True),
        ("2.3", "2.3.1", True),
        ("99.102.0.19", "99.102.1.0", True),
        ("99.102.0.19", "99.102.0.0", False),
        ("99.102.0.19", "99.102.0.19", False),
    ],
)
def test_ps_version_less_than(version, other, expected):
    assert (primitive.PSVersion(version) < other) == expected


@pytest.mark.parametrize(
    "version, other, expected",
    [
        ("2.2", "2.3", True),
        ("2.3", "2.3", True),
        ("2.4", "2.3", False),
        ("3.0", "2.3", False),
        ("1.0", "2.3", True),
        ("2.3.0", "2.3", False),
        ("2.3.1", "2.3", False),
        ("2.3", "2.3.0", True),
        ("2.3", "2.3.1", True),
        ("99.102.0.19", "99.102.1.0", True),
        ("99.102.0.19", "99.102.0.0", False),
        ("99.102.0.19", "99.102.0.19", True),
    ],
)
def test_ps_version_less_or_equal(version, other, expected):
    assert (primitive.PSVersion(version) <= other) == expected


def test_ps_version_compare_invalid():
    expected = re.escape("'>=' not supported between instances of 'PSVersion' and 'int")
    with pytest.raises(TypeError, match=expected):
        primitive.PSVersion("1.0") >= 1


def test_ps_secure_string():
    ps_value = primitive.PSSecureString(COMPLEX_STRING)
    assert str(ps_value) == "System.Security.SecureString"
    assert repr(ps_value) == "PSSecureString()"
    assert ps_value.decrypt() == COMPLEX_STRING

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert (
        actual
        == "<SS>dAByAGUAYgBsAGUAIABjAGwAZQBmAAoAIABfAHgAMAAwADAAMABfACAAXwBYADAAMAAwADAAXwAgADTYHt0gAGMAYQBmAOkAIAA82A==</SS>"
    )

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSSecureString)
    assert not isinstance(actual, str)
    assert str(actual) == (
        "dAByAGUAYgBsAGUAIABjAGwAZQBmAAoAIABfAHgAMAAwADAAMABfACAAXwBYADAAMAAwADAAXwAgADTYHt0gAGMAYQBmAOkAIAA82A=="
    )
    assert actual.PSTypeNames == ["System.Security.SecureString", "System.Object"]

    dec_actual = actual.decrypt()
    assert isinstance(dec_actual, primitive.PSString)
    assert dec_actual == COMPLEX_STRING


def test_ps_secure_string_with_properties():
    ps_value = primitive.PSSecureString("abc")
    ps_value.PSObject.extended_properties.append(PSNoteProperty("Test Property"))
    ps_value["Test Property"] = primitive.PSSecureString("abc")

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == f'<Obj RefId="0"><SS>YQBiAGMA</SS><MS><SS N="Test Property">YQBiAGMA</SS></MS></Obj>'

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSSecureString)
    assert not isinstance(actual, str)
    assert str(actual) == "YQBiAGMA"
    assert isinstance(actual["Test Property"], primitive.PSSecureString)
    assert not isinstance(actual["Test Property"], str)
    assert str(actual["Test Property"]) == "YQBiAGMA"
    assert actual.PSTypeNames == ["System.Security.SecureString", "System.Object"]

    dec_actual = actual.decrypt()
    assert isinstance(dec_actual, primitive.PSString)
    assert dec_actual == "abc"

    dec_actual = actual["Test Property"].decrypt()
    assert isinstance(dec_actual, primitive.PSString)
    assert dec_actual == "abc"


def test_ps_secure_string_already_encrypted():
    ps_value = primitive.PSSecureString("YQBiAGMA", FakeCryptoProvider())
    assert isinstance(ps_value, primitive.PSSecureString)
    assert str(ps_value) == "YQBiAGMA"
    assert ps_value.decrypt() == "abc"

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert actual == f"<SS>YQBiAGMA</SS>"

    actual = deserialize(element)
    assert isinstance(actual, primitive.PSSecureString)
    assert not isinstance(actual, str)
    assert str(ps_value) == "YQBiAGMA"
    assert ps_value.decrypt() == "abc"
